# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Storage implementations for caching results.
"""

from __future__ import annotations

from typing import Mapping, Type
from urllib.parse import urlsplit

from pallas.storage.base import NotFoundError, Storage, UnsupportedURIError
from pallas.storage.filesystem import FileSystemStorage
from pallas.storage.memory import MemoryStorage
from pallas.storage.s3 import S3Storage

__all__ = [
    "Storage",
    "UnsupportedURIError",
    "NotFoundError",
    "storage_from_uri",
]


STORAGE_REGISTRY: Mapping[str, Type[Storage]] = {
    "memory": MemoryStorage,
    "file": FileSystemStorage,
    "s3": S3Storage,
}


def storage_from_uri(uri: str, *, default_scheme: str = "file") -> Storage:
    """
    Construct a storage instance from a string URI.
    """
    scheme, *rest = urlsplit(uri, scheme=default_scheme)
    try:
        cls = STORAGE_REGISTRY[scheme]
    except KeyError:
        raise UnsupportedURIError(f"Unknown scheme: {scheme}.") from None
    return cls.from_uri(uri)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from contextlib import contextmanager

import pytest

from pallas import Athena
from pallas.exceptions import AthenaQueryError
from pallas.testing import FakeProxy


def fake_sleep(seconds):
    raise KeyboardInterrupt


@pytest.fixture(name="fake")
def fake_fixture():
    fake = FakeProxy()
    fake.state = "RUNNING"
    return fake


@pytest.fixture(name="athena")
def athena_fixture(fake):
    orig_sleep = time.sleep
    time.sleep = fake_sleep
    athena = Athena(fake)
    athena.kill_on_interrupt = True
    yield athena
    time.sleep = orig_sleep


@contextmanager
def no_keyboard_interrupt():
    """
    Convert KeyboardInterrupt to test failures.

    When KeyboardInterrupt is not caught by AthenaKillOnInterruptWrapper,
    do not want to terminate the pytest process.
    """
    try:
        yield
    except KeyboardInterrupt:
        raise AssertionError("Unexpected KeyboardInterrupt")


class TestAthenaKillOnInterrupt:
    def test_execute_kill_on_interrupt(self, athena, fake):
        with no_keyboard_interrupt():
            with pytest.raises(AthenaQueryError):
                athena.execute("SELECT 1")
            assert fake.request_log == [
                "StartQueryExecution",
                "GetQueryExecution",
                "StopQueryExecution",
                "GetQueryExecution",
            ]

    def test_execute_do_not_kill_on_interrupt(self, athena, fake):
        athena.kill_on_interrupt = False
        with pytest.raises(KeyboardInterrupt):
            athena.execute("SELECT 1")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
        ]

    def test_get_results_kill_on_interrupt(self, athena, fake):
        with no_keyboard_interrupt():
            query = athena.submit("SELECT 1")
            fake.request_log.clear()
            with pytest.raises(AthenaQueryError):
                query.get_results()
            assert fake.request_log == [
                "GetQueryExecution",
                "StopQueryExecution",
                "GetQueryExecution",
            ]

    def test_get_results_do_not_kill_on_interrupt(self, athena, fake):
        athena.kill_on_interrupt = False
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        with pytest.raises(KeyboardInterrupt):
            query.get_results()
        assert fake.request_log == [
            "GetQueryExecution",
        ]

    def test_join_kill_on_interrupt(self, athena, fake):
        with no_keyboard_interrupt():
            query = athena.submit("SELECT 1")
            fake.request_log.clear()
            with pytest.raises(AthenaQueryError):
                query.join()
            assert fake.request_log == [
                "GetQueryExecution",
                "StopQueryExecution",
                "GetQueryExecution",
            ]

    def test_join_do_not_kill_on_interrupt(self, athena, fake):
        athena.kill_on_interrupt = False
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        with pytest.raises(KeyboardInterrupt):
            query.join()
            assert fake.request_log == [
                "GetQueryExecution",
            ]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CSV reading and writing.

Athena uses a custom CSV format:
 - All values are quoted. Quotes in values are doubled.
 - Empty unquoted strings denotes a missing values.

Unfortunately, the format cannot be parsed using Python builtin CSV module.
"""

from __future__ import annotations

import io
import re
from typing import Iterable, Iterator, Optional, Sequence, TextIO, Tuple

CSVValue = Optional[str]
CSVRow = Sequence[CSVValue]


control_re = re.compile(r'\n|,|"')
quote_re = re.compile(r'"')


def _encode_value(value: CSVValue) -> str:
    if value is None:
        return ""
    quoted = value.replace('"', '""')
    return f'"{quoted}"'


def _decode_value(raw: str) -> CSVValue:
    if raw == "":
        return None
    if not (raw.startswith('"') and raw.endswith('"')):
        raise ValueError(f"Value not quoted: {raw}")
    parts = raw[1:-1].split('""')
    if any('"' in part for part in parts):
        raise ValueError(f"Invalid quoting: {raw}")
    return '"'.join(parts)


def _tokenize(stream: TextIO) -> Iterator[Tuple[str, str]]:
    """
    Tokenize CSV input.

    Yields (value, control) pairs, where:
    - Value is raw field values, including all quoting.
    - Control is one of:
      "," (field separator), "\n" (line separator), "" (end of file).

    """
    buffer = ""  # Chunk of text read from the stream.
    pos = 0  # Position in the buffer.
    value = ""  # Last field, possibly read in multiple chunks.
    quoted = False  # True when inside quoted value.
    while True:
        if not pos < len(buffer):
            buffer = stream.read(io.DEFAULT_BUFFER_SIZE)
            pos = 0
            if not buffer:
                break  # End of file
        pattern = quote_re if quoted else control_re
        match = pattern.search(buffer, pos)
        if match:
            start, end = match.span()
            value += buffer[pos:start]
            control = buffer[start:end]
            pos = end
        else:
            value += buffer[pos:]
            control = ""
            pos = len(buffer)
        if control == '"':
            value += control
            quoted = not quoted
        elif control:
            yield value, control
            value = ""
    if quoted:
        raise ValueError("Unterminated quoted value.")
    yield value, ""


def read_csv(stream: TextIO) -> Iterator[CSVRow]:
    """
    Read CSV using format that Athena uses.

    All values are quoted (quotes are doubled).
    Empty unquoted string denotes ``None``.

    :param stream: readable file-like
    :return: sequence of records
    """
    row = []
    for value, control in _tokenize(stream):
        if control:
            row.append(_decode_value(value))
        if control == "\n":
            yield tuple(row)
            row = []
    if row:
        raise ValueError("Missing trailing newline.")


def write_csv(data: Iterable[CSVRow], stream: TextIO) -> None:
    """
    Write CSV using format that Athena uses.

    All values are quoted (quotes are doubled).
    Empty unquoted string denotes ``None``.

    :param data: sequence of records
    :param stream: writable file-like
    """
    for row in data:
        line = ",".join(_encode_value(v) for v in row)
        stream.write(line + "\n")



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
from decimal import Decimal

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from pallas.results import QueryResults

sample_column_names = "id", "name"
sample_column_types = "integer", "varchar"

sample_data = [
    ("1", "foo"),
    ("2", "bar"),
]


sample_results = QueryResults(sample_column_names, sample_column_types, sample_data)


class TestQueryResults:
    def test_repr(self):
        assert repr(sample_results) == (
            "<QueryResults:"
            " 2 results,"
            " column_names=['id', 'name'],"
            " column_types=['integer', 'varchar']"
            ">"
        )

    def test_list(self):
        assert list(sample_results) == [
            {"id": 1, "name": "foo"},
            {"id": 2, "name": "bar"},
        ]

    def test_len(self):
        assert len(sample_results) == 2

    def test_get_item(self):
        assert sample_results[0] == {"id": 1, "name": "foo"}

    def test_get_slice(self):
        assert list(sample_results[:1]) == [{"id": 1, "name": "foo"}]

    def test_get_slice_from_end(self):
        assert list(sample_results[-1:]) == [{"id": 2, "name": "bar"}]

    def test_to_df(self):
        actual = sample_results.to_df()
        expected = pd.DataFrame(
            {
                "id": pd.Series([1, 2], dtype="Int32"),
                "name": pd.Series(["foo", "bar"], dtype="string"),
            }
        )
        assert_frame_equal(actual, expected)

    def test_to_df_custom_dtypes(self):
        name_dtype = pd.CategoricalDtype(["foo", "bar"])
        actual = sample_results.to_df(dtypes={"id": "int64", "name": name_dtype})
        expected = pd.DataFrame(
            {
                "id": pd.Series([1, 2], dtype="int64"),
                "name": pd.Series(["foo", "bar"], dtype=name_dtype),
            }
        )
        assert_frame_equal(actual, expected)

    @pytest.mark.parametrize(
        "column_type,values,series",
        [
            ("unknown", [None], pd.Series([None], dtype="string")),
            (
                "varchar",
                [None, "", "value"],
                pd.Series([None, "", "value"], dtype="string"),
            ),
            (
                "char",
                [None, "", "value"],
                pd.Series([None, "", "value"], dtype="string"),
            ),
            (
                "boolean",
                [None, "false", "true"],
                pd.Series([None, False, True], dtype="boolean"),
            ),
            (
                "tinyint",
                [None, "-128", "0", "127"],
                pd.Series([None, -(2 ** 7), 0, 2 ** 7 - 1], dtype="Int8"),
            ),
            (
                "smallint",
                [None, "-32768", "0", "32767"],
                pd.Series([None, -(2 ** 15), 0, 2 ** 15 - 1], dtype="Int16"),
            ),
            (
                "integer",
                [None, "-2147483648", "0", "2147483647"],
                pd.Series([None, -(2 ** 31), 0, 2 ** 31 - 1], dtype="Int32"),
            ),
            (
                "bigint",
                [None, "-9223372036854775808", "0", "9223372036854775807"],
                pd.Series([None, -(2 ** 63), 0, 2 ** 63 - 1], dtype="Int64"),
            ),
            (
                "float",
                [None, "NaN", "1.2E34", "-Infinity"],
                pd.Series([np.nan, np.nan, 1.2e34, -np.inf], dtype="float32"),
            ),
            (
                "double",
                [None, "NaN", "1.2E34", "-Infinity"],
                pd.Series([np.nan, np.nan, 1.2e34, -np.inf], dtype="float64"),
            ),
            (
                "decimal",
                [None, "0", "0.1"],
                pd.Series([None, Decimal("0"), Decimal("0.1")], dtype="object"),
            ),
            (
                "date",
                [None, "2001-08-22"],
                pd.Series([pd.NaT, dt.datetime(2001, 8, 22)], dtype="datetime64[ns]"),
            ),
            (
                "timestamp",
                [None, "2001-08-22 03:04:05.321"],
                pd.Series(
                    [pd.NaT, dt.datetime(2001, 8, 22, 3, 4, 5, 321000)],
                    dtype="datetime64[ns]",
                ),
            ),
            (
                "varbinary",
                [None, "", "00 ff 00"],
                pd.Series([None, b"", b"\x00\xff\x00"], dtype="object"),
            ),
            (
                "array",
                [None, "[]", "[a, b]"],
                pd.Series([None, [], ["a", "b"]], dtype="object"),
            ),
            (
                "map",
                [None, "{}", "{k1=v1, k2=v2}"],
                pd.Series([None, {}, {"k1": "v1", "k2": "v2"}], dtype="object"),
            ),
            (
                "json",
                [None, "null", "1", '{"a": []}'],
                pd.Series([None, None, 1, {"a": []}], dtype="object"),
            ),
        ],
    )
    def test_to_df_types(self, column_type, values, series):
        data = [(v,) for v in values]
        results = QueryResults(["col"], [column_type], data)
        assert_frame_equal(
            results.to_df(), pd.DataFrame({"col": series}), check_column_type="exact"
        )



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from pallas.storage import NotFoundError, storage_from_uri
from pallas.storage.filesystem import FileSystemStorage
from pallas.storage.memory import MemoryStorage
from pallas.storage.s3 import S3Storage


@pytest.fixture(name="memory_storage")
def memory_storage_fixture(tmp_path):
    return MemoryStorage()


@pytest.fixture(name="file_storage")
def file_storage_fixture(tmp_path):
    return FileSystemStorage(tmp_path / "cache")


@pytest.fixture(name="s3_storage")
def s3_storage_fixture(s3_tmp_uri):
    return S3Storage.from_uri(s3_tmp_uri)


@pytest.fixture(name="storage", params=["memory", "file", "s3"])
def storage_fixture(request):
    """
    Yields all implemented storage implementations.
    """
    yield request.getfixturevalue(f"{request.param}_storage")


class TestStorage:
    def test_get_missing(self, storage):
        with pytest.raises(NotFoundError):
            storage.get("foo")

    def test_set_get(self, storage):
        storage.set("foo", "Hello Foo!")
        assert storage.get("foo") == "Hello Foo!"

    def test_has_missing(self, storage):
        assert not storage.has("foo")

    def test_has_present(self, storage):
        storage.set("foo", "Hello Foo!")
        assert storage.has("foo")

    def test_read_missing(self, storage):
        with pytest.raises(NotFoundError):
            storage.reader("foo")

    def test_read(self, storage):
        storage.set("foo", "Hello Foo,\nHope you are doing well.")
        with storage.reader("foo") as stream:
            assert stream.readline() == "Hello Foo,\n"
            assert stream.readline() == "Hope you are doing well."

    def test_write(self, storage):
        with storage.writer("foo") as stream:
            stream.write("Hello Foo,\n")
            stream.write("Hope you are doing well.")
        assert storage.get("foo") == "Hello Foo,\nHope you are doing well."


class TestFromURI:
    def test_invalid_scheme(self):
        with pytest.raises(ValueError):
            storage_from_uri("unknown://")

    @pytest.mark.parametrize("uri", ["memory:", "memory://"])
    def test_memory_from_uri(self, uri):
        storage = storage_from_uri(uri)
        assert isinstance(storage, MemoryStorage)
        assert str(storage) == storage.uri == "memory:"
        assert repr(storage) == "<MemoryStorage: 'memory:'>"

    @pytest.mark.parametrize(
        "uri",
        ["memory://netloc", "memory:path", "memory:?query=1", "memory:#fragment"],
    )
    def test_invalid_memory_from_uri(self, uri):
        with pytest.raises(ValueError):
            storage_from_uri(uri)

    @pytest.mark.parametrize(
        "uri,base_dir",
        [
            ("/path", "/path"),
            ("path", os.path.abspath("./path")),
            ("./path", os.path.abspath("./path")),
            ("~/path", os.path.expanduser("~/path")),
            ("file:/path", "/path"),
            ("file:path", os.path.abspath("./path")),
            ("file:./path", os.path.abspath("./path")),
            ("file:~/path", os.path.expanduser("~/path")),
            ("file:///path", "/path"),
        ],
    )
    def test_file_from_uri(self, uri, base_dir):
        storage = storage_from_uri(uri)
        assert isinstance(storage, FileSystemStorage)
        assert str(storage) == storage.uri == f"file:{base_dir}/"
        assert repr(storage) == f"<FileSystemStorage: 'file:{base_dir}/'>"
        assert str(storage.base_dir) == base_dir

    @pytest.mark.parametrize(
        "uri",
        [
            "file:",
            "file://",
            "file://netloc",
            "file:/path?query=1",
            "file:/path#fragment",
        ],
    )
    def test_invalid_file_from_uri(self, uri):
        with pytest.raises(ValueError):
            storage_from_uri(uri)

    @pytest.mark.parametrize(
        "uri,bucket,prefix",
        [
            ("s3://bucket", "bucket", ""),
            ("s3://bucket/", "bucket", ""),
            ("s3://bucket/path", "bucket", "path/"),
            ("s3://bucket/path/", "bucket", "path/"),
        ],
    )
    def test_s3_from_uri(self, uri, bucket, prefix):
        storage = storage_from_uri(uri)
        assert isinstance(storage, S3Storage)
        assert str(storage) == storage.uri == f"s3://{bucket}/{prefix}"
        assert repr(storage) == f"<S3Storage: 's3://{bucket}/{prefix}'>"
        assert str(storage.bucket) == bucket
        assert str(storage.prefix) == prefix

    @pytest.mark.parametrize(
        "uri",
        [
            "s3:",
            "s3://",
            "s3:path",
            "s3:/path",
            "s3:///path",
            "s3://bucket?query=1",
            "s3://bucket#fragment",
        ],
    )
    def test_invalid_s3_from_uri(self, uri):
        with pytest.raises(ValueError):
            storage_from_uri(uri)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import secrets
from urllib.parse import urlsplit

import boto3
import pytest


@pytest.fixture(name="region_name", scope="session")
def region_name_fixture():
    # Do not raise or skip tests if region is not defined.
    # Region can be defined in ~/.aws/config.
    return os.environ.get("PALLAS_TEST_REGION")


@pytest.fixture(name="athena_database", scope="session")
def athena_database_fixture():
    """
    Athena database.

    Tests depending on this fixture are skipped
    if the PALLAS_TEST_ATHENA_DATABASE environment variable is not defined.
    """
    database = os.environ.get("PALLAS_TEST_ATHENA_DATABASE")
    if not database:
        pytest.skip("PALLAS_TEST_ATHENA_DATABASE not defined.")
    return database


@pytest.fixture(name="athena_workgroup", scope="session")
def athena_workgroup_fixture():
    """
    Athena workgroup.
    """
    return os.environ.get("PALLAS_TEST_ATHENA_WORKGROUP")


def _s3_recursive_delete(uri):
    scheme, netloc, path, query, fragment = urlsplit(uri)
    assert scheme == "s3"
    assert query == fragment == ""
    if path and not path.endswith("/"):
        path += "/"
    bucket = boto3.resource("s3").Bucket(netloc)
    for item in bucket.objects.filter(Prefix=path):
        item.delete()


@pytest.fixture(name="s3_session_tmp_uri", scope="session")
def s3_session_tmp_uri_fixture():
    """
    Base URI of a temporary S3 locations.

    Performs cleanup at the end of the test session.

    Tests depending on this fixture are skipped
    if the PALLAS_TEST_S3_TMP environment variable is not defined.
    """
    base_uri = os.environ.get("PALLAS_TEST_S3_TMP")
    if not base_uri:
        pytest.skip("PALLAS_TEST_S3_TMP not defined.")
    if base_uri and not base_uri.endswith("/"):
        base_uri += "/"
    token = secrets.token_hex(4)  # Unique path allows parallel test runs.
    uri = base_uri + f"test-pallas-{token}"
    yield uri
    _s3_recursive_delete(uri)


@pytest.fixture(name="s3_tmp_uri")
def s3_tmp_uri_fixture(s3_session_tmp_uri):
    """
    URI of a temporary S3 location than can be used for testing.

    A unique URI is generated for each test.
    Cleanup is performed at once at the end of the test session.

    Tests depending on this fixture are skipped
    if the PALLAS_TEST_S3_TMP environment variable is not defined.
    """
    if s3_session_tmp_uri and not s3_session_tmp_uri.endswith("/"):
        s3_session_tmp_uri += "/"
    return s3_session_tmp_uri + secrets.token_hex(8)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Caching of Athena queries.
"""

from __future__ import annotations

import hashlib
import logging
from typing import Optional
from urllib.parse import urlencode

from pallas.results import QueryResults
from pallas.storage import NotFoundError, Storage, storage_from_uri

logger = logging.getLogger("pallas")


def _get_execution_key(database: Optional[str], sql: str) -> str:
    parts = {}
    if database is not None:
        parts["database"] = database
    parts["sql"] = sql
    encoded = urlencode(parts).encode("utf-8")
    digest = hashlib.sha256(encoded).hexdigest()
    return f"query-{digest}"


def _get_results_key(execution_id: str) -> str:
    return f"results-{execution_id}"


def _load_execution_id(
    storage: Optional[Storage], database: Optional[str], sql: str
) -> Optional[str]:
    if storage is None:
        return None
    key = _get_execution_key(database, sql)
    try:
        execution_id = storage.get(key)
    except NotFoundError:
        return None
    logger.info(
        f"Query execution loaded from cache {storage}{key}:"
        f" QueryExecutionId={execution_id!r}"
    )
    return execution_id


def _save_execution_id(
    storage: Optional[Storage], database: Optional[str], sql: str, execution_id: str
) -> None:
    if storage is None:
        return
    key = _get_execution_key(database, sql)
    storage.set(key, execution_id)
    logger.info(
        f"Query execution saved to cache {storage}{key}:"
        f" QueryExecutionId={execution_id!r}"
    )


def _has_results(storage: Optional[Storage], execution_id: str) -> bool:
    if storage is None:
        return False
    key = _get_results_key(execution_id)
    if not storage.has(key):
        return False
    logger.info(
        f"Query results are available in cache {storage}{key}:"
        f" QueryExecutionId={execution_id!r}"
    )
    return True


def _load_results(
    storage: Optional[Storage], execution_id: str
) -> Optional[QueryResults]:
    if storage is None:
        return None
    key = _get_results_key(execution_id)
    try:
        with storage.reader(key) as stream:
            results = QueryResults.load(stream)
    except NotFoundError:
        return None
    logger.info(
        f"Query results loaded from cache {storage}{key}:"
        f" QueryExecutionId={execution_id!r}: {len(results)} rows"
    )
    return results


def _save_results(
    storage: Optional[Storage], execution_id: str, results: QueryResults
) -> None:
    if storage is None:
        return
    key = _get_results_key(execution_id)
    with storage.writer(key) as stream:
        results.save(stream)
    logger.info(
        f"Query results saved to cache {storage}{key}:"
        f" QueryExecutionId={execution_id!r}: {len(results)} rows"
    )


class AthenaCache:
    """
    Caches queries and its results.

    Athena always stores results in S3, so it is possible
    to retrieve past results without manually copying data.

    This class can hold a reference to two instances of cache storage:

    - local, which caches both query execution IDs and query results
    - remote, which cache query execution IDs only.

    It is possible to configure one the backends, both of them,
    or none of them.

    Queries cached in the local storage can be executed without
    an internet connection.
    Queries cached in the remote storage are not executed twice,
    but results have to be downloaded from AWS.

    In theory, it is possible to use remote backend for the local
    cache (or vice versa), but we assume that the local cache
    is actually stored locally

    Instance of this class is returned by the :attr:`.Athena.cache` property.
    It can be updated to reconfigure the caching.
    """

    local_storage: Optional[Storage] = None
    remote_storage: Optional[Storage] = None

    #: Can be set to False to disable caching completely.
    #:
    #: Can be update to enable or disable the caching.
    enabled: bool = True

    #: Can be set to False to disable reading the cache.
    #:
    #: Can be update to reconfigure the caching.
    read: bool = True

    #: Can be set to False to disable writing the cache.
    #:
    #: Can be update to reconfigure the caching.
    write: bool = True

    @property
    def local(self) -> Optional[str]:
        """
        URI of storage for local cache.

        Can be updated to reconfigure the caching.
        """
        if self.local_storage is None:
            return None
        return self.local_storage.uri

    @local.setter
    def local(self, uri: Optional[str]) -> None:
        if uri is None:
            self.local_storage = None
        else:
            self.local_storage = storage_from_uri(uri)

    @property
    def remote(self) -> Optional[str]:
        """
        URI of storage for remote cache.

        Can be updated to reconfigure the caching.
        """
        if self.remote_storage is None:
            return None
        return self.remote_storage.uri

    @remote.setter
    def remote(self, uri: Optional[str]) -> None:
        if uri is None:
            self.remote_storage = None
        else:
            self.remote_storage = storage_from_uri(uri)

    def load_execution_id(self, database: Optional[str], sql: str) -> Optional[str]:
        """
        Retrieve cached query execution ID for the given SQL.

        Looks into both the local and the remote storage.
        """
        if not (self.enabled and self.read):
            return None
        execution_id = _load_execution_id(self.local_storage, database, sql)
        if execution_id is not None:
            return execution_id
        execution_id = _load_execution_id(self.remote_storage, database, sql)
        if execution_id is not None:
            _save_execution_id(self.local_storage, database, sql, execution_id)
            return execution_id
        return None

    def save_execution_id(
        self, database: Optional[str], sql: str, execution_id: str
    ) -> None:
        """
        Store cached query execution ID for the given SQL.

        Updates both the local and the remote storage.
        """
        if not (self.enabled and self.write):
            return
        _save_execution_id(self.remote_storage, database, sql, execution_id)
        _save_execution_id(self.local_storage, database, sql, execution_id)

    def has_results(self, execution_id: str) -> bool:
        """
        Checks whether results are cached for the given execution ID.

        Looks into the local storage only.
        """
        if not (self.enabled and self.read):
            return False
        return _has_results(self.local_storage, execution_id)

    def load_results(self, execution_id: str) -> Optional[QueryResults]:
        """
        Retrieve cached results for the given execution ID.

        Looks into the local storage only.
        """
        if not (self.enabled and self.read):
            return None
        return _load_results(self.local_storage, execution_id)

    def save_results(self, execution_id: str, results: QueryResults) -> None:
        """
        Store cached results for the given SQL.

        Updates the local storage only.
        """
        if not (self.enabled and self.write):
            return
        _save_results(self.local_storage, execution_id, results)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import NoReturn


class UnavailableModule:
    """
    Fake module that can be imported but raises when used.
    """

    def __init__(self, module: str) -> None:
        self.__module = module

    def __getattr__(self, item: str) -> NoReturn:
        raise RuntimeError(f"{self.__module} cannot be imported.")


try:
    import numpy
except ImportError:
    numpy = UnavailableModule("numpy")

try:
    import pandas
except ImportError:
    pandas = UnavailableModule("pandas")



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from pallas.utils import Fibonacci, truncate_str


def test_fibonacci_wo_max_value():
    generator = Fibonacci()
    items = itertools.islice(generator, 12)
    assert list(items) == [1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144]


def test_fibonacci_w_max_value():
    generator = Fibonacci(max_value=60)
    items = itertools.islice(generator, 12)
    assert list(items) == [1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 60, 60]


def test_truncate_str_short():
    assert truncate_str("Hello world!") == "Hello world!"


def test_truncate_str_long():
    s = "Hello, " + 20 * "hello, " + "world!"
    assert truncate_str(s) == (
        "Hello, hello, hello, hello, hello, hello, hello,"
        " hell...lo, hello, hello, world!"
    )



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
from decimal import Decimal

import numpy as np
import pytest

from pallas import Athena
from pallas.sql import is_select, normalize_sql, substitute_parameters


@pytest.mark.parametrize(
    "value,quoted",
    [
        (None, "NULL"),
        ("", "''"),
        ("hello", "'hello'"),
        ("'", "''''"),
        (42, "42"),
        (np.int64(42), "42"),
        (3.14, "3.14"),
        (float("nan"), "nan()"),
        (float("inf"), "infinity()"),
        (float("-inf"), "-infinity()"),
        (np.float64(3.14), "3.14"),
        (np.float64("nan"), "nan()"),
        (np.float64("inf"), "infinity()"),
        (np.float64("-inf"), "-infinity()"),
        (np.nan, "nan()"),
        (np.inf, "infinity()"),
        (-np.inf, "-infinity()"),
        (Decimal("0.1"), "DECIMAL '0.1'"),
        (dt.date(2001, 8, 22), "DATE '2001-08-22'"),
        (
            dt.datetime(2001, 8, 22, 3, 4, 5, 321000),
            "TIMESTAMP '2001-08-22 03:04:05.321'",
        ),
        (b"\x00\xff\x00", "from_base64('AP8A')"),
    ],
)
def test_quote_value(value, quoted):
    assert Athena.quote(value) == quoted


@pytest.mark.parametrize("value", [object(), (), [], {}])
def test_quote_invalid(value):
    with pytest.raises(TypeError):
        Athena.quote(value)


class TestSubstituteParameters:
    def test_substitute_none(self):
        sql = substitute_parameters("SELECT * FROM t WHERE x LIKE 'A%'")
        assert sql == "SELECT * FROM t WHERE x LIKE 'A%'"

    def test_substitute_sequence(self):
        sql = substitute_parameters(
            "SELECT %s, %s FROM t WHERE x LIKE 'A%%'", [1, "foo"]
        )
        assert sql == "SELECT 1, 'foo' FROM t WHERE x LIKE 'A%'"

    def test_substitute_mapping(self):
        sql = substitute_parameters(
            "SELECT %(id)s, %(name)s FROM t WHERE x LIKE 'A%%'",
            {"id": 1, "name": "foo"},
        )
        assert sql == "SELECT 1, 'foo' FROM t WHERE x LIKE 'A%'"

    @pytest.mark.parametrize("value", [1, "foo", object()])
    def test_invalid(self, value):
        with pytest.raises(TypeError):
            substitute_parameters("SELECT %s", value)


class TestIsSelect:
    def test_select(self):
        assert is_select("SELECT 1")

    def test_select_lowercase(self):
        assert is_select("select 1")

    def test_with_select(self):
        assert is_select("WITH (...) AS t SELECT ...")

    def test_with_select_lowercase(self):
        assert is_select("with (...) AS t select ...")

    def test_insert(self):
        assert not is_select("INSERT ... AS SELECT")

    def test_create(self):
        assert not is_select("CREATE TABLE AS ... SELECT")

    def test_select_without_whitesplace(self):
        assert is_select("SELECT*FROM ...")

    def test_single_line_comments(self):
        assert is_select(
            """
            -- Comment 1
            -- Comment 2
            SELECT
        """
        )

    def test_multi_line_comments(self):
        assert is_select(
            """
            /*
            Comment 1
            Comment 2
            */
            SELECT
        """
        )

    def test_multi_line_comment_escaped(self):
        assert is_select(
            r"""
            /* *\/ */
            SELECT
        """
        )


NORMALIZED_SQL = """\
SELECT
    c1, c2
FROM
    t\
"""


class TestNormalizeSQL:
    def test_dedent(self):
        sql = normalize_sql(
            """
            SELECT
                c1, c2
            FROM
                t
        """
        )
        assert sql == NORMALIZED_SQL

    def test_trailing_whitespace(self):
        sql = normalize_sql(
            """\N{space}
            SELECT\N{space}\N{space}\N{space}
                c1, c2\N{space}
            FROM\N{space}
                t\N{space}
        """
        )
        assert sql == NORMALIZED_SQL

    def test_empty_lines(self):
        # Leading and trailing new lines are removed.
        # Other new lines are normalized to LF
        sql = normalize_sql(
            """\n\n\n
            SELECT\n\N{space}\N{space}\N{space}\n\r\n
                c1, c2
            FROM
                t\n\n\n
        """
        )
        assert sql == NORMALIZED_SQL.replace("SELECT", "SELECT\n\n\n")



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pathlib
from typing import TextIO, Union
from urllib.parse import urlsplit

from pallas.storage.base import NotFoundError, Storage, UnsupportedURIError


def file_parse_uri(uri: str) -> str:
    """
    Parse path from file URI.
    """
    scheme, netloc, path, query, fragment = urlsplit(uri, scheme="file")
    if scheme != "file":
        raise UnsupportedURIError("Not a file scheme.")
    if netloc:
        raise UnsupportedURIError("The scheme does not support hostname.")
    if query or fragment:
        raise UnsupportedURIError("The scheme does not support query or fragment.")
    if not path:
        raise UnsupportedURIError("Path is empty.")
    return path


class FileSystemStorage(Storage):
    """
    Storage implementation storing data on a local filesystem.
    """

    _initialized: bool = False
    _base_dir: pathlib.Path

    def __init__(self, base_dir: Union[str, pathlib.Path]) -> None:
        self._base_dir = pathlib.Path(base_dir).expanduser().absolute()

    @classmethod
    def from_uri(cls, uri: str) -> FileSystemStorage:
        path = file_parse_uri(uri)
        return cls(path)

    @property
    def uri(self) -> str:
        return f"file:{self._base_dir}/"

    @property
    def base_dir(self) -> pathlib.Path:
        return self._base_dir

    def get(self, key: str) -> str:
        try:
            return self._get_file(key).read_text(encoding="utf-8")
        except FileNotFoundError:
            raise NotFoundError(key) from None

    def set(self, key: str, value: str) -> None:
        self._get_file(key).write_text(value, encoding="utf-8")

    def has(self, key: str) -> bool:
        return self._get_file(key).exists()

    def reader(self, key: str) -> TextIO:
        try:
            return self._get_file(key).open("r", encoding="utf-8", newline="")
        except FileNotFoundError:
            raise NotFoundError(key) from None

    def writer(self, key: str) -> TextIO:
        return self._get_file(key).open("w", encoding="utf-8", newline="")

    def _get_file(self, key: str) -> pathlib.Path:
        if not self._initialized:
            # Create cache directory when first used.
            self._base_dir.mkdir(exist_ok=True, parents=True)
            self._initialized = True
        return self.base_dir / key



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Conversions from strings returned by Athena to Python types.
"""

from __future__ import annotations

import datetime as dt
import json
from abc import ABCMeta, abstractmethod
from decimal import Decimal
from typing import Dict, Generic, Iterable, List, Optional, Sequence, TypeVar

from pallas._compat import numpy as np
from pallas._compat import pandas as pd

T_co = TypeVar("T_co", covariant=True)


def _pd_array(values: Sequence[object], *, dtype: object) -> object:
    if dtype == "object":
        # Workaround for ValueError: PandasArray must be 1-dimensional.
        # When all values are lists of same length, Pandas/NumPy think
        # that we are constructing a 2-D array.
        data = np.empty(len(values), dtype="object")
        data[:] = values
    else:
        data = values
    return pd.array(data, dtype=dtype, copy=False)


class Converter(Generic[T_co], metaclass=ABCMeta):
    """
    Convert values returned by Athena to Python types.
    """

    @property
    @abstractmethod
    def dtype(self) -> object:
        """Pandas dtype"""

    def read(self, value: Optional[str]) -> Optional[T_co]:
        """
        Read value returned from Athena.

        Expect a string or ``None`` because optional strings
        are what Athena returns at its API and that is also
        what can be parsed from CSV stored in S3.
        """
        if value is None:
            return None
        return self.read_str(value)

    @abstractmethod
    def read_str(self, value: str) -> T_co:
        """
        Read value from string

        To be implemented in subclasses.
        """

    def read_array(
        self, values: Iterable[Optional[str]], dtype: Optional[object] = None,
    ) -> object:  # Pandas array
        """
        Convert values returned from Athena to Pandas array.

        :param values: Iterable yielding strings and ``None``
        :param dtype: optional Pandas dtype to force
        """
        if dtype is None:
            dtype = self.dtype
        converted = [self.read(value) for value in values]
        return _pd_array(converted, dtype=dtype)


class TextConverter(Converter[str]):
    @property
    def dtype(self) -> object:
        return "string"

    def read_str(self, value: str) -> str:
        return value


class BooleanConverter(Converter[bool]):
    @property
    def dtype(self) -> object:
        return "boolean"

    def read_str(self, value: str) -> bool:
        return {"true": True, "false": False}[value]


class IntConverter(Converter[int]):
    def __init__(self, size: int) -> None:
        self._size = size

    @property
    def dtype(self) -> object:
        return f"Int{self._size}"

    def read_str(self, value: str) -> int:
        return int(value)


class FloatConverter(Converter[float]):
    def __init__(self, size: int) -> None:
        self._size = size

    @property
    def dtype(self) -> object:
        return f"float{self._size}"

    def read_str(self, value: str) -> float:
        return float(value)


class DecimalConverter(Converter[Decimal]):
    @property
    def dtype(self) -> object:
        return "object"

    def read_str(self, value: str) -> Decimal:
        return Decimal(value)


class DateConverter(Converter[dt.date]):
    @property
    def dtype(self) -> object:
        return "datetime64[ns]"

    def read_str(self, value: str) -> dt.date:
        return dt.date.fromisoformat(value)


class DateTimeConverter(Converter[dt.datetime]):
    @property
    def dtype(self) -> object:
        return "datetime64[ns]"

    def read_str(self, value: str) -> dt.datetime:
        return dt.datetime.fromisoformat(value)


class BinaryConverter(Converter[bytes]):
    @property
    def dtype(self) -> object:
        return "object"

    def read_str(self, value: str) -> bytes:
        return bytes.fromhex(value)


class ArrayConverter(Converter[List[str]]):
    """
    Parse string returned by Athena to a list.

    Array parsing has multiple limitations because of the
    serialization format that Athena uses:

     - Always returns a list of strings because Athena does
       not send more details about item types.
     - It is not possible to distinguish comma in values from
       an item separator. We assume that values do not contain the comma.
     - We are not able to distinguish an empty array
       and an array with one empty string.
       This converter returns an empty array in that case.

    """

    @property
    def dtype(self) -> object:
        return "object"

    def read_str(self, value: str) -> List[str]:
        if not value.startswith("[") or not value.endswith("]"):
            raise ValueError(f"Invalid array value: {value!r}")
        content = value[1:-1]
        if not content:
            return []
        return content.split(", ")


class MapConverter(Converter[Dict[str, str]]):
    """
    Convert string value returned from Athena to a dictionary.

    Map parsing has multiple limitations because of the
    serialization format that Athena uses:

    - Always returns a mapping from strings to strings because
      Athena does not send more details about item types.
    - It is not possible to distinguish a comma or an equal sign
      in values from control characters.
      We assume that values do not contain the comma or the equal sign.
    """

    @property
    def dtype(self) -> object:
        return "object"

    def read_str(self, value: str) -> Dict[str, str]:
        if not value.startswith("{") or not value.endswith("}"):
            raise ValueError(f"Invalid map value: {value!r}")
        content = value[1:-1]
        if not content:
            return {}
        parts = (part.partition("=") for part in content.split(", "))
        return {k: v for k, _, v in parts}


class JSONConverter(Converter[object]):
    @property
    def dtype(self) -> object:
        return "object"

    def read_str(self, value: str) -> object:
        return json.loads(value)


default_converter = TextConverter()

CONVERTERS: Dict[str, Converter[object]] = {
    "boolean": BooleanConverter(),
    "tinyint": IntConverter(8),
    "smallint": IntConverter(16),
    "integer": IntConverter(32),
    "bigint": IntConverter(64),
    "float": FloatConverter(32),
    "double": FloatConverter(64),
    "decimal": DecimalConverter(),
    "date": DateConverter(),
    "timestamp": DateTimeConverter(),
    "varbinary": BinaryConverter(),
    "array": ArrayConverter(),
    "map": MapConverter(),
    "json": JSONConverter(),
}


def get_converter(column_type: str) -> Converter[object]:
    """
    Return a converter for a column type.

    :param column_type: a column type as reported by Athena
    :return: a converter instance.
    """
    return CONVERTERS.get(column_type, default_converter)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
from abc import ABCMeta, abstractmethod
from typing import TextIO


class UnsupportedURIError(ValueError):
    """
    Storage URI not supported.
    """


class NotFoundError(KeyError):
    """
    Storage item not found.

    Raised when trying to read a non-existent item.
    """


class StorageWriter(io.StringIO):
    """
    Writable stream that saves content to :class:`.Storage` on close.

    Used by the default implementation of the :meth:`.Storage.writer` method.
    """

    def __init__(self, storage: Storage, key: str) -> None:
        super().__init__()
        self._storage = storage
        self._key = key

    def close(self) -> None:
        self._storage.set(self._key, self.getvalue())
        super().close()


class Storage(metaclass=ABCMeta):
    """
    Interface for a simple key-value storage.

    Implementations re used  as cache backends.
    """

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.uri!r}>"

    def __str__(self) -> str:
        return self.uri

    @classmethod
    @abstractmethod
    def from_uri(cls, uri: str) -> Storage:
        """
        Construct an instance from a string URI.
        """

    @property
    @abstractmethod
    def uri(self) -> str:
        """
        Return a string URI describing this storage instance.
        """

    @abstractmethod
    def get(self, key: str) -> str:
        """
        Get value of the given key.

        Provides a simple method for retrieving short values.
        For longer content, the :meth:`.reader` can be better.

        Raises :exception:`.NotFoundError` if the key is not found.
        """

    @abstractmethod
    def set(self, key: str, value: str) -> None:
        """
        Set value of the given key.

        Provides a simple method for storing short values.
        For longer content, use the :meth:`.writer` method.
        """

    @abstractmethod
    def has(self, key: str) -> bool:
        """
        Test presence of the given key.

        In most cases, it is more efficient to try to retrieve a value
        and catch possible exception than to check presence first.
        """

    def reader(self, key: str) -> TextIO:
        """
        Return a file-like object opened for reading.

        The default implementation returns a in-memory stream
        with the whole content read from storage,
        but subclasses can provide a more efficient method.

        Raises :exception:`.NotFoundError` if the key is not found.
        """
        return io.StringIO(self.get(key))

    def writer(self, key: str) -> TextIO:
        """
        Return a file-like object opened for writing.

        The default implementation buffers written content in memory
        and stores it at once on close,
        but subclasses can provide a more efficient method.
        """
        return StorageWriter(self, key)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pallas import Athena
from pallas.storage.memory import MemoryStorage
from pallas.testing import FakeProxy

FAKE_DATA = [("1", "foo", None)]
ANOTHER_FAKE_DATA = [("2", "bar", None)]


@pytest.fixture(name="storage")
def storage_fixture():
    return MemoryStorage()


@pytest.fixture(name="fake")
def fake_athena_fixture():
    """
    Athena mock decorated by caching wrapper.
    """
    fake = FakeProxy()
    fake.column_names = "id", "name", "value"
    fake.column_types = "integer", "varchar", "double"
    fake.data = FAKE_DATA
    return fake


@pytest.fixture(name="local_athena")
def local_athena_fixture(fake, storage):
    """
    Athena client caching locally.

    In local mode, both query execution IDs and results are cached.
    """
    athena = Athena(fake)
    athena.cache.local_storage = storage
    return athena


@pytest.fixture(name="remote_athena")
def remote_athena_fixture(fake, storage):
    """
    Athena client caching remotely.

    In remote mode, only query execution IDs are cached.
    """
    athena = Athena(fake)
    athena.cache.remote_storage = storage
    return athena


@pytest.fixture(name="full_athena")
def full_athena_fixture(fake, storage):
    """
    Athena client caching both locally and remotely.
    """
    athena = Athena(fake)
    athena.cache.local_storage = storage
    athena.cache.remote_storage = MemoryStorage()
    return athena


@pytest.fixture(name="athena", params=["local", "remote", "full"])
def athena_fixture(request):
    """
    Yields caching wrapper in both remote and local mode.
    """
    return request.getfixturevalue(f"{request.param}_athena")


def assert_query_results(results):
    assert list(results) == [{"id": 1, "name": "foo", "value": None}]


def assert_another_query_results(results):
    assert list(results) == [{"id": 2, "name": "bar", "value": None}]


class TestAthenaCache:

    # Test execute method

    def test_execute_one_query(self, athena, fake):
        """Test execution of a query not in cache."""
        results = athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(results)

    def test_local_execute_one_query_cache_size(self, local_athena, storage):
        """Test that query ID and results are written to cache."""
        local_athena.execute("SELECT 1 id, 'foo' name")
        assert storage.size() == 2

    def test_remote_execute_one_query_cache_size(self, remote_athena, storage):
        """Test that query ID is written to cache."""
        remote_athena.execute("SELECT 1 id, 'foo' name")
        assert storage.size() == 1

    def test_full_execute_one_query_cache_size(self, full_athena):
        """Test that query ID and results are written to cache."""
        full_athena.execute("SELECT 1 id, 'foo' name")
        assert full_athena.cache.local_storage.size() == 2
        assert full_athena.cache.remote_storage.size() == 1

    def test_execute_one_query_not_select(self, athena, fake, storage):
        """Test execution of a query that should not be cached."""
        results = athena.execute("CREATE TABLE ...")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(results)
        assert storage.size() == 0

    def test_local_execute_second_query(self, local_athena, fake, storage):
        """Test execution of a query in cache."""
        local_athena.execute("SELECT 1 id, 'foo' name")  # fill cache
        fake.request_log.clear()
        results = local_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == []
        assert_query_results(results)
        assert storage.size() == 2

    def test_remote_execute_second_query(self, remote_athena, fake, storage):
        """Test execution of a query in cache."""
        remote_athena.execute("SELECT 1 id, 'foo' name")  # fill cache
        fake.request_log.clear()
        results = remote_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(results)
        assert storage.size() == 1

    def test_full_execute_second_query(self, full_athena, fake):
        """Test execution of a query in cache."""
        full_athena.execute("SELECT 1 id, 'foo' name")  # fill cache
        fake.request_log.clear()
        results = full_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == []
        assert_query_results(results)
        assert full_athena.cache.local_storage.size() == 2
        assert full_athena.cache.remote_storage.size() == 1

    def test_execute_second_query_not_select(self, athena, fake, storage):
        """Test that only SELECT queries are cached."""
        athena.execute("CREATE TABLE ...")
        fake.request_log.clear()
        results = athena.execute("CREATE TABLE ...")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(results)
        assert storage.size() == 0

    def test_execute_second_query_different_sql(self, athena, fake, storage):
        """Test that cache is unique to a query."""
        athena.execute("SELECT 1 id, 'foo' name")  # fill cache
        fake.request_log.clear()
        fake.data = ANOTHER_FAKE_DATA
        results = athena.execute("SELECT 2 id, 'bar' name")  # different SQL
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_another_query_results(results)

    def test_execute_second_query_different_database(self, athena, fake, storage):
        """Test that cache is unique to a database."""
        athena.execute("SELECT 1 id, 'foo' name")  # fill cache
        fake.request_log.clear()
        athena.database = "other"
        results = athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(results)

    # Test athena.submit() method

    def test_submit_one_query(self, athena, fake, storage):
        """Test that the caching wrapper submits a query if not in cache."""
        athena.submit("SELECT 1 id, 'foo' name")
        assert fake.request_log == ["StartQueryExecution"]
        assert storage.size() == 1

    def test_submit_one_query_not_select(self, athena, fake, storage):
        """Test that the caching wrapper submits a query if not in cache."""
        athena.submit("CREATE TABLE ...")
        assert fake.request_log == ["StartQueryExecution"]
        assert storage.size() == 0

    def test_submit_second_query(self, athena, fake, storage):
        """Test that one query is submitted only once."""
        athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        athena.submit("SELECT 1 id, 'foo' name")
        assert fake.request_log == []
        assert storage.size() == 1

    def test_submit_second_query_not_select(self, athena, fake, storage):
        """Test that only SELECT queries are cached."""
        athena.submit("CREATE TABLE ...")
        fake.request_log.clear()
        athena.submit("CREATE TABLE ...")
        assert fake.request_log == ["StartQueryExecution"]
        assert storage.size() == 0

    def test_submit_second_query_different_sql(self, athena, fake, storage):
        """Test that cache is unique to a query."""
        athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        fake.data = ANOTHER_FAKE_DATA
        athena.submit("SELECT 2 id, 'bar' name")
        assert fake.request_log == ["StartQueryExecution"]
        assert storage.size() == 2

    # Test athena.get_query() method

    def test_get_uncached_query_get_results(self, athena, fake, storage):
        """Test getting query not in cache."""
        athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query_results = athena.get_query("query-1").get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]
        assert_query_results(query_results)

    def test_local_uncached_query_get_results_cache_size(self, local_athena, storage):
        """Test that results are not cached."""
        local_athena.submit("SELECT 1 id, 'foo' name")
        local_athena.get_query("query-1").get_results()
        assert storage.size() == 2

    def test_remote_uncached_query_get_results_cache_size(self, remote_athena, storage):
        """Test that results are not cached."""
        remote_athena.submit("SELECT 1 id, 'foo' name")
        remote_athena.get_query("query-1").get_results()
        assert storage.size() == 1

    def test_local_get_cached_query_get_results(self, local_athena, fake, storage):
        """Test getting query in cache when results are cached."""
        local_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query_results = local_athena.get_query("query-1").get_results()
        assert fake.request_log == []
        assert_query_results(query_results)
        assert storage.size() == 2

    def test_remote_get_cached_query_get_results(self, remote_athena, fake, storage):
        """Test getting query in cache when results are not cached."""
        remote_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query_results = remote_athena.get_query("query-1").get_results()
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(query_results)
        assert storage.size() == 1

    def test_get_query_get_results_not_select(self, athena, storage):
        """Test that results are cached for SELECT queries onlys."""
        athena.submit("CREATE TABLE ...")
        athena.get_query("query-1").get_results()
        assert storage.size() == 0

    # Test query.get_info() method

    def test_query_info(self, athena, fake):
        """Test obtaining information about a query."""
        query = athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        info = query.get_info()
        assert info.execution_id == "query-1"
        assert info.sql == "SELECT 1 id, 'foo' name"
        assert fake.request_log == ["GetQueryExecution"]

    # Test query.get_results() method

    def test_get_results_one_query(self, athena, fake, storage):
        """Test getting results query not in cache."""
        query = athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query_results = query.get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]
        assert_query_results(query_results)

    def test_local_get_second_results_one_query(self, local_athena, fake):
        """Test getting results twice results cached."""
        query = local_athena.submit("SELECT 1 id, 'foo' name")
        query.get_results()
        fake.request_log.clear()
        query_results = query.get_results()
        assert fake.request_log == []
        assert_query_results(query_results)

    def test_remote_get_second_results_one_query(self, remote_athena, fake):
        """Test getting results twice results not cached."""
        query = remote_athena.submit("SELECT 1 id, 'foo' name")
        query.get_results()
        fake.request_log.clear()
        query_results = query.get_results()
        assert fake.request_log == [
            "GetQueryResults",
        ]
        assert_query_results(query_results)

    def test_local_get_results_second_query_same_sql(self, local_athena, fake):
        """Test getting results query in cache results cached."""
        local_athena.execute("SELECT 1 id, 'foo' name")
        second_query = local_athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        second_query_results = second_query.get_results()
        assert fake.request_log == []
        assert_query_results(second_query_results)

    def test_remote_get_results_second_query_same_sql(self, remote_athena, fake):
        """Test getting results query in cache results not cached."""
        remote_athena.execute("SELECT 1 id, 'foo' name")
        second_query = remote_athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        second_query_results = second_query.get_results()
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(second_query_results)

    def test_get_results_second_query_different_sql(self, athena, fake):
        """Test that cache is unique to a query."""
        athena.execute("SELECT 1 id, 'foo' name")
        fake.data = ANOTHER_FAKE_DATA
        second_query = athena.submit("SELECT 2 id, 'bar' name")
        fake.request_log.clear()
        second_query_results = second_query.get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]
        assert_another_query_results(second_query_results)

    def test_local_get_uncached_results_second_query_same_sql(self, local_athena, fake):
        """Test that the second query downloads data if the first does not."""
        local_athena.submit("SELECT 1 id, 'foo' name")  # Does not download results.
        second_query = local_athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        second_query_results = second_query.get_results()
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert_query_results(second_query_results)

    def test_local_get_cached_results_first_query_same_sql(self, local_athena, fake):
        """Test that first query can use cache from the second query."""
        first_query = local_athena.submit("SELECT 1 id, 'foo' name")
        local_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        first_query_results = first_query.get_results()
        assert fake.request_log == []
        assert_query_results(first_query_results)

    # Test query.kill() method

    def test_kill(self, athena, fake):
        query = athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query.kill()
        assert fake.request_log == ["StopQueryExecution"]

    # Test query.join() method

    def test_join_one_query(self, athena, fake):
        """Test waiting for a query not cached."""
        query = athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        query.join()
        assert fake.request_log == ["GetQueryExecution"]

    def test_local_join_second_query_query_same_sql(self, local_athena, fake):
        """Test waiting for a query cached results not cached."""
        local_athena.execute("SELECT 1 id, 'foo' name")
        second_query = local_athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        second_query.join()
        assert fake.request_log == []

    def test_remote_join_second_query_query_same_sql(self, remote_athena, fake):
        """Test waiting for a query cached results not cached."""
        remote_athena.execute("SELECT 1 id, 'foo' name")
        second_query = remote_athena.submit("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        second_query.join()
        assert fake.request_log == ["GetQueryExecution"]

    def test_join_second_query_query_different_sql(self, athena, fake):
        """Test that cache is unique to a query."""
        athena.execute("SELECT 1 id, 'foo' name")
        fake.data = ANOTHER_FAKE_DATA
        second_query = athena.submit("SELECT 2 id, 'bar' name")
        fake.request_log.clear()
        second_query.join()
        assert fake.request_log == ["GetQueryExecution"]

    # Test configuration

    def test_cache_disabled_first_query(self, athena, fake, storage):
        athena.cache.enabled = False
        athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert storage.size() == 0

    def test_cache_disabled_second_query(self, athena, fake):
        athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        athena.cache.enabled = False
        athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]

    def test_local_cache_read_disabled_first_query(self, local_athena, fake, storage):
        local_athena.cache.read = False
        local_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert storage.size() == 2

    def test_remote_cache_read_disabled_first_query(self, remote_athena, fake, storage):
        remote_athena.cache.read = False
        remote_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert storage.size() == 1

    def test_cache_read_disabled_second_query(self, athena, fake):
        athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        athena.cache.read = False
        athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]

    def test_cache_write_disabled_first_query(self, athena, fake, storage):
        athena.cache.write = False
        athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert storage.size() == 0

    def test_local_cache_write_disabled_second_query(self, local_athena, fake):
        local_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        local_athena.cache.write = False
        local_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == []

    def test_remote_cache_write_disabled_second_query(self, remote_athena, fake):
        remote_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        remote_athena.cache.write = False
        remote_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_cached_remotely_only(self, full_athena, fake):
        """
        Test that remote cache is used to populate local cache.
        """
        # Another client populates remote cache.
        another_athena = Athena(fake)
        another_athena.cache.remote_storage = full_athena.cache.remote_storage
        another_athena.execute("SELECT 1 id, 'foo' name")
        fake.request_log.clear()
        # First execution uses remote cache
        full_athena.execute("SELECT 1 id, 'foo' name")
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
        ]
        assert full_athena.cache.local_storage.size() == 2



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Methods for setting up Athena clients.
"""

from __future__ import annotations

import os
from typing import Mapping, Optional

from pallas.client import Athena
from pallas.proxies import Boto3Proxy


def setup(
    *,
    region: Optional[str] = None,
    database: Optional[str] = None,
    workgroup: Optional[str] = None,
    output_location: Optional[str] = None,
    cache_local: Optional[str] = None,
    cache_remote: Optional[str] = None,
    normalize: bool = True,
    kill_on_interrupt: bool = True,
) -> Athena:
    """
    Setup an :class:`.Athena` client.

    All configuration options can be given to this method,
    but many of them can be overridden after the client is constructed.

    :param region: an AWS region.
        By default, region from AWS config (``~/.aws/config``) is used.
    :param database: a name of Athena database.
        Can be overridden in SQL.
    :param workgroup: a name of Athena workgroup.
        Workgroup can set resource limits or override output location.
        Defaults to the Athena default workgroup.
    :param output_location: an output location at S3 for query results.
        Optional if an output location is specified for the *workgroup*.
    :param cache_local: an URI of a local cache.
        Both results and query execution IDs are stored in the local cache.
    :param cache_remote: an URI of a remote cache.
        Query execution IDs without results are stored in the remote cache.
    :param normalize: whether to normalize queries before execution.
    :param kill_on_interrupt: whether to kill queries on KeyboardInterrupt.
    :return: a new instance of Athena client
    """
    athena = Athena(Boto3Proxy(region=region))
    if cache_local is not None:
        athena.cache.local = cache_local
    if cache_remote is not None:
        athena.cache.remote = cache_remote
    athena.database = database
    athena.workgroup = workgroup
    athena.output_location = output_location
    athena.normalize = normalize
    athena.kill_on_interrupt = kill_on_interrupt
    return athena


def environ_setup(
    environ: Optional[Mapping[str, str]] = None, *, prefix: str = "PALLAS"
) -> Athena:
    """
    Setup an :class:`.Athena` client from environment variables.

    Reads the following environment variables: ::

        export PALLAS_REGION=
        export PALLAS_DATABASE=
        export PALLAS_WORKGROUP=
        export PALLAS_OUTPUT_LOCATION=
        export PALLAS_NORMALIZE=true
        export PALLAS_KILL_ON_INTERRUPT=true
        export PALLAS_CACHE_REMOTE=$PALLAS_OUTPUT_LOCATION
        export PALLAS_CACHE_LOCAL=~/Notebooks/.cache/

    Configuration from the environment variables can be overridden
    after the client is constructed.

    :param environ: A mapping object representing the string environment.
        Defaults to ``os.environ``.
    :param prefix: A prefix of environment variables
    :return: a new instance of Athena client
    """
    if environ is None:
        environ = os.environ
    config = _EnvironConfig(environ, prefix)
    return setup(
        database=config.get_str("DATABASE"),
        workgroup=config.get_str("WORKGROUP"),
        output_location=config.get_str("OUTPUT_LOCATION"),
        region=config.get_str("REGION"),
        cache_remote=config.get_str("CACHE_REMOTE"),
        cache_local=config.get_str("CACHE_LOCAL"),
        normalize=config.get_bool("NORMALIZE", True),
        kill_on_interrupt=config.get_bool("KILL_ON_INTERRUPT", True),
    )


class _EnvironConfig:
    def __init__(self, environ: Mapping[str, str], prefix: str) -> None:
        self._environ = environ
        self._prefix = prefix

    def get_str(self, key: str, default: Optional[str] = None) -> Optional[str]:
        v = self._get(key)
        if not v:
            return default
        return v

    def get_bool(self, key: str, default: bool = False) -> bool:
        v = self._get(key)
        if not v:
            return default
        v = v.lower()
        if v in ("1", "true", "on", "yes"):
            return True
        if v in ("0", "false", "off", "no"):
            return False
        raise ValueError(f"{self._prefix}_{key}: invalid boolean value: {v}")

    def _get(self, key: str) -> str:
        v = self._environ.get(f"{self._prefix}_{key}", "")
        if not isinstance(v, str):
            # Avoid unexpected behaviour when somebody passes something
            # like {"PALLAS_KILL_ON_INTERRUPT": False} to environ.
            raise TypeError("Environ values must be a string.")
        return v



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Dict
from urllib.parse import urlsplit

from pallas.storage.base import NotFoundError, Storage, UnsupportedURIError


def memory_parse_uri(uri: str) -> None:
    """
    Check that memory URI is valid.
    """
    scheme, netloc, path, query, fragment = urlsplit(uri, scheme="memory")
    if scheme != "memory":
        raise UnsupportedURIError("Not a memory scheme.")
    if netloc or path or query or fragment:
        raise UnsupportedURIError("Superfluous component.")


class MemoryStorage(Storage):
    """
    Storage implementation storing data in memory.

    Useful mainly for testing.
    """

    _data: Dict[str, str]

    def __init__(self) -> None:
        self._data = {}

    @classmethod
    def from_uri(cls, uri: str) -> MemoryStorage:
        memory_parse_uri(uri)
        return cls()

    @property
    def uri(self) -> str:
        return "memory:"

    def get(self, key: str) -> str:
        try:
            return self._data[key]
        except KeyError:
            raise NotFoundError(key) from None

    def set(self, key: str, value: str) -> None:
        self._data[key] = value

    def has(self, key: str) -> bool:
        return key in self._data

    def size(self) -> int:
        return len(self._data)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Proxies to AWS Athena APIs.

The proxies are internal classes by the :class:`.Athena` client
to issue requests to AWS.
"""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Mapping, Optional, Sequence, cast

import boto3

from pallas.csv import read_csv
from pallas.info import QueryInfo
from pallas.results import QueryResults
from pallas.storage.s3 import s3_parse_uri, s3_wrap_body
from pallas.utils import truncate_str

logger = logging.getLogger("pallas")


ColumnNames = Sequence[str]
ColumnTypes = Sequence[str]
Row = Sequence[Optional[str]]


class AthenaProxy(metaclass=ABCMeta):
    """
    Proxy to AWS Athena.

    This is an internal interface that is used by the :class:`.Athena` client.
    The :class:`.Boto3Proxy` implementation will be used in most cases,
    but it can be replaced by :class:`.FakeProxy` for testing.
    """

    @abstractmethod
    def start_query_execution(
        self,
        sql: str,
        *,
        database: Optional[str] = None,
        workgroup: Optional[str] = None,
        output_location: Optional[str] = None,
    ) -> str:
        """
        Submit a query.

        :param sql: an SQL query to be executed
        :param database: a name of Athena database to be queried
        :param workgroup: a name of Athena workgroup
        :param output_location: URI of output location on S3
        :return: execution_id
        """

    @abstractmethod
    def get_query_execution(self, execution_id: str) -> QueryInfo:
        """
        Retrieve information about a query execution.

        Returns a status of the query with other information.
        """

    @abstractmethod
    def get_query_results(self, info: QueryInfo) -> QueryResults:
        """
        Retrieve results of a query execution.

        Waits until the query execution finishes and downloads results.
        """

    @abstractmethod
    def stop_query_execution(self, execution_id: str) -> None:
        """
        Kill a query execution.
        """


class Boto3Proxy(AthenaProxy):
    """
    Proxy to AWS Athena using the boto3 library.

    This is an internal class that is used by the :class:`.Athena` client.
    It can be replaced by :class:`.FakeProxy` for testing.
    """

    _athena_client: Any  # boto3 Athena client
    _s3_client: Any  # boto3 S3 client

    def __init__(
        self,
        *,
        region: Optional[str] = None,
        athena_client: Optional[Any] = None,
        s3_client: Optional[Any] = None,
    ) -> None:
        if athena_client is None:
            athena_client = boto3.client("athena", region_name=region)
        if s3_client is None:
            s3_client = boto3.client("s3")
        self._athena_client = athena_client
        self._s3_client = s3_client

    def start_query_execution(
        self,
        sql: str,
        database: Optional[str] = None,
        workgroup: Optional[str] = None,
        output_location: Optional[str] = None,
    ) -> str:
        params: Dict[str, Any] = dict(QueryString=sql)
        if database is not None:
            params.update(QueryExecutionContext={"Database": database})
        if workgroup is not None:
            params.update(WorkGroup=workgroup)
        if output_location is not None:
            params.update(ResultConfiguration={"OutputLocation": output_location})
        logger.info(f"Athena StartQueryExecution: QueryString={truncate_str(sql)!r}")
        response = self._athena_client.start_query_execution(**params)
        execution_id = cast(str, response["QueryExecutionId"])
        logger.info(f"Athena QueryExecutionId={execution_id!r} started.")
        return execution_id

    def get_query_execution(self, execution_id: str) -> QueryInfo:
        logger.info(f"Athena GetQueryExecution: QueryExecutionId={execution_id!r}")
        response = self._athena_client.get_query_execution(
            QueryExecutionId=execution_id
        )
        info = QueryInfo(response["QueryExecution"])
        logger.info(f"Athena QueryExecution: {info}")
        return info

    def get_query_results(self, info: QueryInfo) -> QueryResults:
        execution_id = info.execution_id
        params = dict(QueryExecutionId=execution_id)
        logger.info(f"Athena GetQueryResults: QueryExecutionId={execution_id!r}")
        response = self._athena_client.get_query_results(**params)
        column_names = _read_column_names(response)
        column_types = _read_column_types(response)
        if response.get("NextToken"):
            logger.info("Athena ResultSet paginated. Will download from S3.")
            data = self._download_data(info)
        else:
            data = _read_data(response)
            logger.info(
                f"Athena ResultSet complete: {len(data)} rows (including header)"
            )
        fixed_data = _fix_data(column_names, data)
        return QueryResults(column_names, column_types, fixed_data)

    def stop_query_execution(self, execution_id: str) -> None:
        logger.info(f"Athena StopQueryExecution: QueryExecutionId={execution_id!r}")
        self._athena_client.stop_query_execution(QueryExecutionId=execution_id)

    def _download_data(self, info: QueryInfo) -> Sequence[Row]:
        output_location = info.output_location
        bucket, key = s3_parse_uri(output_location)
        params = dict(Bucket=bucket, Key=key)
        logger.info(f"S3 GetObject:" f" Bucket={bucket!r} Key={key!r}")
        response = self._s3_client.get_object(**params)
        with s3_wrap_body(response["Body"]) as stream:
            data = list(read_csv(stream))
        logger.info(f"S3 Body downloaded: {len(data)} rows (including header)")
        return data


def _fix_data(column_names: ColumnNames, data: Sequence[Row]) -> Sequence[Row]:
    """
    Fix malformed data returned from Athena.

    Queries executed by Presto (typically queries with SELECT)
    repeat column names in the first row of data,
    so we have to remove them.

    Queries by Hive (typically queries with DESCRIBE)
    do not repeat column names, but all columns are combined to one.

    Try to fix both of the above problems here.
    """
    if data and data[0] == column_names:
        # DQL, SELECT statements executed by Presto
        data = data[1:]
    elif all(len(row) == 1 for row in data) and len(column_names) > 1:
        # DCL, DESCRIBE statements executed by Hive
        values = (row[0] for row in data if row[0] is not None)
        data = [v.split("\t", maxsplit=len(column_names) - 1) for v in values]
    return data


def _read_column_names(response: Mapping[str, Any]) -> ColumnNames:
    column_info = response["ResultSet"]["ResultSetMetadata"]["ColumnInfo"]
    return tuple(column["Name"] for column in column_info)


def _read_column_types(response: Mapping[str, Any]) -> ColumnTypes:
    column_info = response["ResultSet"]["ResultSetMetadata"]["ColumnInfo"]
    return tuple(column["Type"] for column in column_info)


def _read_data(response: Mapping[str, Any]) -> Sequence[Row]:
    rows = response["ResultSet"]["Rows"]
    return [tuple(item.get("VarCharValue") for item in row["Data"]) for row in rows]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Encapsulation of results returned from Athena.
"""

from __future__ import annotations

from typing import Dict, Mapping, Optional, Sequence, TextIO, Union, cast, overload

from pallas._compat import pandas as pd
from pallas.conversions import Converter, get_converter
from pallas.csv import read_csv, write_csv

QueryRecord = Dict[str, object]


class QueryResults(Sequence[QueryRecord]):
    """
    Collection of Athena query results.

    Implements a list-like interface for accessing individual records.
    Alternatively, can be converted to :class:`pandas.DataFrame`
    using the :meth:`.to_df` method.
    """

    _column_names: Sequence[str]
    _column_types: Sequence[str]
    _data: Sequence[Sequence[Optional[str]]]

    def __init__(
        self,
        column_names: Sequence[str],
        column_types: Sequence[str],
        data: Sequence[Sequence[Optional[str]]],
    ) -> None:
        self._column_names = column_names
        self._column_types = column_types
        self._data = data

    def __repr__(self) -> str:
        parts = [
            f"{len(self)} results",
            f"column_names={self.column_names!r}",
            f"column_types={self.column_types!r}",
        ]
        return f"<{type(self).__name__}: {', '.join(parts)}>"

    @overload
    def __getitem__(self, index: int) -> QueryRecord:
        ...

    @overload  # noqa: F811
    def __getitem__(self, index: slice) -> Sequence[QueryRecord]:
        ...

    def __getitem__(  # noqa: F811
        self, index: Union[int, slice]
    ) -> Union[QueryRecord, Sequence[QueryRecord]]:
        """
        Return one result or slice of results.

        Records are returned as mappings from column names to values.
        """
        if isinstance(index, slice):
            return QueryResults(self.column_names, self.column_types, self._data[index])
        row = self._data[index]
        return {
            cn: converter.read(v)
            for cn, converter, v in zip(self.column_names, self._converters, row)
        }

    def __len__(self) -> int:
        """
        Return count of this results.
        """
        return len(self._data)

    @classmethod
    def load(cls, stream: TextIO) -> QueryResults:
        """
        Deserialize results from a text stream.
        """
        reader = read_csv(stream)
        column_names = next(reader)
        column_types = next(reader)
        data = list(reader)
        if any(v is None for v in column_names):
            raise ValueError("Missing column name")
        if any(v is None for v in column_types):
            raise ValueError("Missing column type")
        column_names = cast(Sequence[str], column_names)
        column_types = cast(Sequence[str], column_types)
        return cls(column_names, column_types, data)

    def save(self, stream: TextIO) -> None:
        """
        Serialize results to a text stream.
        """
        write_csv([self._column_names, self._column_types], stream)
        write_csv(self._data, stream)

    @property
    def column_names(self) -> Sequence[str]:
        """List of column names."""
        return list(self._column_names)

    @property
    def column_types(self) -> Sequence[str]:
        """List of column types."""
        return list(self._column_types)

    @property
    def _converters(self) -> Sequence[Converter[object]]:
        return list(map(get_converter, self.column_types))

    def to_df(self, dtypes: Optional[Mapping[str, object]] = None) -> pd.DataFrame:
        """
        Convert this results to :class:`pandas.DataFrame`.
        """
        if pd is None:
            raise RuntimeError("Pandas cannot be imported.")
        if dtypes is None:
            dtypes = {}
        frame_data = {}
        for i, (name, converter) in enumerate(zip(self.column_names, self._converters)):
            values = (row[i] for row in self._data)
            frame_data[name] = converter.read_array(values, dtype=dtypes.get(name))
        return pd.DataFrame(frame_data, copy=False)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pallas.assembly import environ_setup, setup


class TestSetup:
    def test_default(self):
        athena = setup(output_location="s3://example-output/")
        assert athena.output_location == "s3://example-output/"
        assert athena.normalize is True
        assert athena.kill_on_interrupt is True

    def test_default_from_env(self):
        athena = environ_setup(
            environ={"PALLAS_OUTPUT_LOCATION": "s3://example-output/"}
        )
        assert athena.output_location == "s3://example-output/"
        assert athena.normalize is True
        assert athena.kill_on_interrupt is True
        assert athena.cache.local is None
        assert athena.cache.remote is None

    def test_do_not_normalize(self):
        athena = setup(normalize=False)
        assert athena.normalize is False

    def test_do_not_normalize_from_env(self):
        athena = environ_setup(environ={"PALLAS_NORMALIZE": "0"})
        assert athena.normalize is False

    def test_do_not_kill_on_interrupt(self):
        athena = setup(kill_on_interrupt=False)
        assert athena.kill_on_interrupt is False

    def test_do_not_kill_on_interrupt_from_env(self):
        athena = environ_setup(environ={"PALLAS_KILL_ON_INTERRUPT": "0"})
        assert athena.kill_on_interrupt is False

    def test_cache_local(self):
        athena = setup(cache_local="/path")
        assert athena.cache.local == "file:/path/"

    def test_cache_local_from_env(self):
        athena = environ_setup(environ={"PALLAS_CACHE_LOCAL": "/path"})
        assert athena.cache.local == "file:/path/"

    def test_cache_remote(self):
        athena = setup(cache_remote="s3://bucket/path/")
        assert athena.cache.remote == "s3://bucket/path/"

    def test_cache_remote_from_env(self):
        athena = environ_setup(environ={"PALLAS_CACHE_REMOTE": "s3://bucket/path/"})
        assert athena.cache.remote == "s3://bucket/path/"

    def test_cache_local_and_remote(self):
        athena = setup(cache_remote="s3://bucket/path/", cache_local="/path")
        assert athena.cache.local == "file:/path/"
        assert athena.cache.remote == "s3://bucket/path/"

    def test_cache_local_and_remote_from_env(self):
        athena = environ_setup(
            environ={
                "PALLAS_CACHE_REMOTE": "s3://bucket/path/",
                "PALLAS_CACHE_LOCAL": "/path",
            }
        )
        assert athena.cache.local == "file:/path/"
        assert athena.cache.remote == "s3://bucket/path/"



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import time
from typing import Iterable, Optional

from pallas.caching import AthenaCache
from pallas.info import QueryInfo
from pallas.proxies import AthenaProxy
from pallas.results import QueryResults
from pallas.sql import (
    PARAMETERS,
    is_select,
    normalize_sql,
    quote,
    substitute_parameters,
)
from pallas.utils import Fibonacci


class Query:
    """
    Athena query

    Provides access to one query execution.
    It can be used to monitor status of the query results
    or retrieving results when the execution finishes.

    Instances of this class are returned by :meth:`Athena.submit`
    and :meth:`Athena.get_query` methods.
    You should not need to create this class directly.

    :param execution_id: Athena query execution ID.
    :param proxy: an internal proxy to execute queries
    :param cache: a cache instance
    """

    #: Delays in seconds between for checking query status.
    backoff: Iterable[int] = Fibonacci(max_value=60)

    #: Whether to kill this query on KeyboardInterrupt
    #:
    #: Initially set to :attr:`Athena.kill_on_interrupt`.
    kill_on_interrupt: bool = False

    _execution_id: str
    _proxy: AthenaProxy
    _cache: AthenaCache

    _info: Optional[QueryInfo] = None

    def __init__(
        self, execution_id: str, *, proxy: AthenaProxy, cache: AthenaCache,
    ) -> None:
        self._execution_id = execution_id
        self._proxy = proxy
        self._cache = cache

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: execution_id={self.execution_id!r}>"

    @property
    def execution_id(self) -> str:
        """
        Athena query execution ID.

        This ID can be used to retrieve this query later using
        the :meth:`.Athena.get_query()` method.
        """
        return self._execution_id

    def get_info(self) -> QueryInfo:
        """
        Retrieve information about this query execution.

        Returns a status of this query with other information.
        """
        # Query info is cached if the query finished and cannot change.
        if self._info is not None:
            return self._info
        info = self._proxy.get_query_execution(self._execution_id)
        if info.finished:
            self._info = info
        return info

    def get_results(self) -> QueryResults:
        """
        Download results of this query execution.

        Cached results can be returned, if the caching was configured.
        Only SELECT queries are cached.

        Waits until this query execution finishes and downloads results.
        Raises :class:`.AthenaQueryError` if the query failed.
        """
        # When a user calls athena.get_query(execution_id).get_results(),
        # we have to look into the cache without knowing what SQL was executed,
        # so whether the query is cacheable.
        results = self._cache.load_results(self._execution_id)
        if results is not None:
            return results
        self.join()
        info = self.get_info()
        results = self._proxy.get_query_results(info)
        should_cache = is_select(info.sql)
        if should_cache:
            self._cache.save_results(self._execution_id, results)
        return results

    def kill(self) -> None:
        """
        Kill this query execution.

        This is a non-blocking operation.
        It does not wait until the query is killed.
        """
        self._proxy.stop_query_execution(self._execution_id)

    def join(self) -> None:
        """
        Wait until this query execution finishes.

        Raises :class:`.AthenaQueryError` if the query failed.
        """
        # When we have results locally, exit without calling any AWS API.
        if self._cache.has_results(self._execution_id):
            return
        for delay in self.backoff:
            info = self.get_info()
            if info.finished:
                info.check()
                break
            try:
                time.sleep(delay)
            except KeyboardInterrupt:
                if not self.kill_on_interrupt:
                    raise
                # Catch only the first KeyboardInterrupt
                self.kill_on_interrupt = False
                self.kill()
                self.join()


class Athena:
    """
    Athena client.

    Provides methods to execute SQL queries in AWS Athena,
    with an optional caching and other helpers.

    Can be used as a blocking or a non-blocking client.

    Use :func:`.setup` or :func:`.environ_setup` to construct
    this class without touching Pallas internals.

    :param proxy: an internal proxy to execute queries
    """

    quote = staticmethod(quote)

    #: Name of Athena database to be be queried.
    #:
    #: Can be overridden in SQL.
    database: Optional[str] = None

    #: Name of Athena workgroup.
    #:
    #: Workgroup can set resource limits or override output location.
    #: When None, defaults to the Athena default workgroup.
    workgroup: Optional[str] = None

    #: URI of output location on S3.
    #:
    #: Optional if an output location is specified for :attr:`.workgroup`.
    output_location: Optional[str] = None

    #: Whether to normalize queries before execution.
    normalize: bool = True

    #: Whether to kill queries on KeyboardInterrupt
    kill_on_interrupt: bool = True

    _proxy: AthenaProxy
    _cache: AthenaCache

    def __init__(self, proxy: AthenaProxy) -> None:

        self._proxy = proxy
        self._cache = AthenaCache()

    def __repr__(self) -> str:
        parts = [
            f"database={self.database!r}",
            f"workgroup={self.workgroup!r}",
            f"output_location={self.output_location!r}",
        ]
        return f"<{type(self).__name__}: {', '.join(parts)}>"

    @property
    def proxy(self) -> AthenaProxy:
        return self._proxy

    @property
    def cache(self) -> AthenaCache:
        """
        Cache implementation.

        It is possible to update properties of the :attr:`.cache`
        attribute to reconfigure caching in place.

        Alternatively, the :meth:`.using` method can apply
        a new configuration without affecting an existing instance.

        :rtype: :class:`.AthenaCache`
        """
        return self._cache

    def using(
        self,
        *,
        database: Optional[str] = None,
        workgroup: Optional[str] = None,
        output_location: Optional[str] = None,
        normalize: Optional[bool] = None,
        kill_on_interrupt: Optional[bool] = None,
        cache_enabled: Optional[bool] = None,
        cache_read: Optional[bool] = None,
        cache_write: Optional[bool] = None,
    ) -> Athena:
        """
        Crate a new instance with an updated configuration.

        This method can be useful if you need to override a configuration
        for one query, but you do not want to affect future queries.

        :param database: name of Athena database to be be queried.
        :param workgroup: name of Athena workgroup.
        :param output_location: URI of output location on S3.
        :param normalize: whether to normalize queries before execution.
        :param kill_on_interrupt: whether to kill queries on KeyboardInterrupt
        :param cache_enabled: whether a cache should be used.
        :param cache_read: whether a cache should be read.
        :param cache_write: whether a cache should be written.
        :return: an updated copy of this client
        """
        other = copy.copy(self)
        other._cache = copy.copy(self._cache)
        if database is not None:
            other.database = database
        if workgroup is not None:
            other.workgroup = workgroup
        if output_location is not None:
            other.output_location = output_location
        if normalize is not None:
            other.normalize = normalize
        if kill_on_interrupt is not None:
            other.kill_on_interrupt = kill_on_interrupt
        if cache_enabled is not None:
            other._cache.enabled = cache_enabled
        if cache_read is not None:
            other._cache.read = cache_read
        if cache_write is not None:
            other._cache.write = cache_write
        return other

    def execute(
        self,
        operation: str,
        parameters: PARAMETERS = None,
    ) -> QueryResults:
        """
        Execute a query and return results.

        This is a blocking method that waits until the query finishes.

        Cached results or results from an existing query can be returned,
        if the caching was configured. Only SELECT queries are cached.

        Raises :class:`.AthenaQueryError` if the query fails.

        :param operation: an SQL query to be executed
            Can contain ``%s`` or ``%(key)s`` placeholders for substitution
            by *parameters*.
        :param parameters: parameters to substitute in *operation*.
            All substitute parameters are quoted appropriately.
            See the :meth:`.quote` method for a supported parameter types.
        :type parameters: Union[None, Tuple[SQL_SCALAR, ...], Mapping[str, SQL_SCALAR]]
        :return: query results
        """
        return self.submit(operation, parameters).get_results()

    def submit(
        self,
        operation: str,
        parameters: PARAMETERS = None,
    ) -> Query:
        """
        Submit a query and return.

        This is a non-blocking method that starts a query and returns.
        Returns a :class:`Query` instance for monitoring query execution
        and downloading results later.

        An existing query can be returned, if the caching was configured.
        Only SELECT queries are cached.

        :param operation: an SQL query to be executed
            Can contain ``%s`` or ``%(key)s`` placeholders for substitution
            by *parameters*.
        :param parameters: parameters to substitute in *operation*.
            All substitute parameters are quoted appropriately.
            See the :meth:`.quote` method for a supported parameter types.
        :type parameters: Union[None, Tuple[SQL_SCALAR, ...], Mapping[str, SQL_SCALAR]]
        :return: a query instance
        """
        sql = self._get_sql(operation, parameters)
        should_cache = is_select(sql)
        if should_cache:
            execution_id = self._cache.load_execution_id(self.database, sql)
            if execution_id is not None:
                return self.get_query(execution_id)
        execution_id = self._proxy.start_query_execution(
            sql,
            database=self.database,
            workgroup=self.workgroup,
            output_location=self.output_location,
        )
        if should_cache:
            self._cache.save_execution_id(self.database, sql, execution_id)
        return self.get_query(execution_id)

    def get_query(self, execution_id: str) -> Query:
        """
        Get a previously submitted query execution.

        This method can be used to retrieve a query executed in the past.
        Because Athena stores results in S3 and does not delete them by default,
        it is possible to download results until they are manually deleted.

        :param execution_id: an Athena query execution ID.
        :return: a query instance
        """
        query = Query(execution_id, proxy=self._proxy, cache=self._cache)
        query.kill_on_interrupt = self.kill_on_interrupt
        return query

    def _get_sql(self, operation: str, parameters: PARAMETERS) -> str:
        sql = substitute_parameters(operation, parameters)
        if self.normalize:
            sql = normalize_sql(sql)
        return sql



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Encapsulation of information returned by Athena GetQueryExecution method.
"""

from __future__ import annotations

import datetime as dt
from typing import Any, Mapping, Optional, cast

from pallas.exceptions import get_error

unit_prefixes = ["k", "M", "G", "T"]


def format_price(v: float) -> str:
    """
    Format price in dollars.
    """
    if v > 1:
        return "$%.2f" % v
    return "%.2f¢" % (100 * v)


def format_size(v: float) -> str:
    """
    Format size in bytes.

    This function assumes that 1kB = 1000B.
    """
    if v < 1000:
        return f"{v:.0f}B"
    for prefix in unit_prefixes:
        v /= 1000
        if v < 1000:
            break
    return f"{v:.2f}{prefix}B"


def format_time(v: dt.timedelta) -> str:
    """
    Format time.
    """
    if v < dt.timedelta(minutes=1):
        s = v.total_seconds()
        return f"{s:.1f}s"
    m, s = divmod(v.seconds, 60)
    return f"{m}min {s}s"


class QueryInfo:
    """
    Information about query execution.

    Instances are returned by the :meth:`.Query.get_info` method.

    :param data: data returned by Athena GetQueryExecution API method.
    """

    def __init__(self, data: Mapping[str, Any]) -> None:
        self._data = data

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {str(self)!r}>"

    def __str__(self) -> str:
        """
        Return summary info about the query execution.

        This is included in logs generated by the Athena client.
        """
        return (
            f"{self.state}, "
            f"scanned {format_size(self.scanned_bytes)} "
            f"in {format_time(self.execution_time)}, "
            f"approx. price {format_price(self.approx_price)}"
        )

    @property
    def execution_id(self) -> str:
        """ID od the query execution."""
        return cast(str, self._data["QueryExecutionId"])

    @property
    def sql(self) -> str:
        """SQL query executed."""
        rv = self._data["Query"]
        return cast(str, rv)

    @property
    def output_location(self) -> str:
        """URI of output location on S3 for the query"""
        rv = self._data["ResultConfiguration"].get("OutputLocation")
        return cast(str, rv)

    @property
    def database(self) -> Optional[str]:
        """Name of database."""
        rv = self._data["QueryExecutionContext"].get("Database")
        return cast(Optional[str], rv)

    @property
    def finished(self) -> bool:
        """Whether the query execution finished."""
        return self.state in ("SUCCEEDED", "FAILED", "CANCELLED")

    @property
    def succeeded(self) -> bool:
        """Whether the query execution finished successfully."""
        return self.state == "SUCCEEDED"

    @property
    def state(self) -> str:
        """State of the query execution."""
        rv = self._data["Status"]["State"]
        return cast(str, rv)

    @property
    def state_reason(self) -> Optional[str]:
        """Reason of the state of the query execution."""
        rv = self._data["Status"].get("StateChangeReason")
        return cast(Optional[str], rv)

    @property
    def scanned_bytes(self) -> int:
        """Data scanned by Athena."""
        rv = self._data["Statistics"].get("DataScannedInBytes", 0)
        return cast(int, rv)

    @property
    def execution_time(self) -> dt.timedelta:
        """Time spent by Athena."""
        milliseconds = self._data["Statistics"].get("TotalExecutionTimeInMillis", 0)
        return dt.timedelta(milliseconds=milliseconds)

    @property
    def approx_price(self) -> float:
        price_per_tb = 5  # https://aws.amazon.com/athena/pricing/
        return price_per_tb * self.scanned_bytes / 10 ** 12

    def check(self) -> None:
        """
        Raises :class:`.AthenaQueryError` (or its subclass) if the query failed.

        Does not raise if the query is still running.
        """
        if self.finished and not self.succeeded:
            raise get_error(self.state, self.state_reason)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pallas import Athena
from pallas.testing import FakeProxy


@pytest.fixture(name="fake")
def fake_fixture():
    return FakeProxy()


@pytest.fixture(name="athena")
def athena_fixture(fake):
    athena = Athena(fake)
    athena.database = "example_database"
    athena.workgroup = "example-workgroup"
    athena.output_location = "s3://example-output/"
    return athena


class TestAthena:
    def test_repr(self, athena):
        assert repr(athena) == (
            "<Athena: database='example_database', workgroup='example-workgroup',"
            " output_location='s3://example-output/'>"
        )

    def test_submit(self, athena, fake):
        athena.submit("SELECT 1")
        assert fake.request_log == ["StartQueryExecution"]

    def test_get_query(self, athena, fake):
        orig_query = athena.submit("SELECT 1")
        fake.request_log.clear()
        athena.get_query(orig_query.execution_id)
        assert fake.request_log == []

    def test_execute(self, athena, fake):
        athena.execute("SELECT 1")
        assert fake.request_log == [
            "StartQueryExecution",
            "GetQueryExecution",
            "GetQueryResults",
        ]

    def test_submit_w_parameters(self, athena):
        query = athena.submit("SELECT %s", ["Joe's"])
        assert query.get_info().sql == "SELECT 'Joe''s'"

    def test_execute_w_parameters(self, athena):
        athena.execute("SELECT %s", ["Joe's"])
        assert athena.get_query("query-1").get_info().sql == "SELECT 'Joe''s'"

    def test_normalize(self, athena):
        query = athena.submit(" SELECT 1 ")
        assert query.get_info().sql == "SELECT 1"

    def test_do_not_normalize(self, athena):
        athena.normalize = False
        query = athena.submit(" SELECT 1 ")
        assert query.get_info().sql == " SELECT 1 "

    def test_using(self, athena):
        other_athena = athena.using(
            database="other_database",
            workgroup="other-workgroup",
            output_location="s3://other-output/",
            kill_on_interrupt=False,
            normalize=False,
        )
        assert other_athena.database == "other_database"
        assert other_athena.workgroup == "other-workgroup"
        assert other_athena.output_location == "s3://other-output/"
        assert other_athena.kill_on_interrupt is False
        assert other_athena.normalize is False

    def test_using_cache(self, athena):
        other_athena = athena.using(cache_enabled=False)
        assert other_athena.cache.enabled is False
        assert other_athena.cache.read is True
        assert other_athena.cache.write is True
        other_athena = athena.using(cache_read=False)
        assert other_athena.cache.enabled is True
        assert other_athena.cache.read is False
        assert other_athena.cache.write is True
        other_athena = athena.using(cache_write=False)
        assert other_athena.cache.enabled is True
        assert other_athena.cache.read is True
        assert other_athena.cache.write is False


class TestQuery:
    def test_repr(self, athena):
        query = athena.submit("SELECT 1")
        assert repr(query) == "<Query: execution_id='query-1'>"

    def test_get_info(self, athena, fake):
        """Test that query info is retrieved."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution"]

    def test_get_info_twice(self, athena, fake):
        """Test that query info is cached"""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_info()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution"]

    def test_get_info_twice_not_finished(self, athena, fake):
        """Test that query info is not cached when query not finished."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        fake.state = "RUNNING"
        query.get_info()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution", "GetQueryExecution"]

    def test_join(self, athena, fake):
        """Test that query info is checked."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.join()
        assert fake.request_log == ["GetQueryExecution"]

    def test_join_twice(self, athena, fake):
        """Test that query info is cached."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.join()
        query.join()
        assert fake.request_log == ["GetQueryExecution"]

    def test_get_info_and_join(self, athena, fake):
        """Test that query info is cached."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_info()
        query.join()
        assert fake.request_log == ["GetQueryExecution"]

    def test_join_and_get_info(self, athena, fake):
        """Test that query info is cached."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.join()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution"]

    def test_get_results(self, athena, fake):
        """Test that get_results check query info."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_get_results_twice(self, athena, fake):
        """Test that get_results caches query info but not results."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_results()
        query.get_results()
        assert fake.request_log == [
            "GetQueryExecution",
            "GetQueryResults",
            "GetQueryResults",
        ]

    def test_get_info_and_get_results(self, athena, fake):
        """Test that get_results use cached query info."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_info()
        query.get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_join_and_get_results(self, athena, fake):
        """Test that get_results use cached query info."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.join()
        query.get_results()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_get_results_and_get_info(self, athena, fake):
        """Test that get_results cache query info."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_results()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_get_results_and_join(self, athena, fake):
        """Test that get_results cache query info."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.get_results()
        query.join()
        assert fake.request_log == ["GetQueryExecution", "GetQueryResults"]

    def test_kill(self, athena, fake):
        """Test that query is killed."""
        query = athena.submit("SELECT 1")
        fake.request_log.clear()
        query.kill()
        assert fake.request_log == ["StopQueryExecution"]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
import itertools
import textwrap
from decimal import Decimal

import pytest

from pallas import Athena
from pallas.exceptions import AthenaQueryError
from pallas.proxies import Boto3Proxy


@pytest.fixture
def athena(region_name, athena_database, athena_workgroup, s3_tmp_uri):
    proxy = Boto3Proxy(region=region_name)
    athena = Athena(proxy)
    athena.database = athena_database
    athena.workgroup = athena_workgroup
    athena.output_location = s3_tmp_uri
    return athena


# Should not be to trivial.
# Test must be able to submit a query,
# and check its status or kill it before the query finishes.
EXAMPLE_SQL = textwrap.dedent(
    """\
    SELECT * FROM
        (VALUES 0, 1, 2, 3, 4, 5, 6, 7, 8, 9) AS t1 (v1),
        (VALUES 0, 1, 2, 3, 4, 5, 6, 7, 8, 9) AS t2 (v2),
        (VALUES 0, 1, 2, 3, 4, 5, 6, 7, 8, 9) AS t3 (v3)
    ORDER BY
        v1, v2, v3\
"""
)


class TestBoto3Proxy:
    def test_success(self, athena):
        query = athena.submit(EXAMPLE_SQL)
        # Running
        info = query.get_info()
        assert not info.finished
        assert not info.succeeded
        assert info.database == athena.database
        assert info.sql == EXAMPLE_SQL
        assert info.state in ("QUEUED", "RUNNING")
        # Finished
        query.join()
        info = query.get_info()
        assert info.finished
        assert info.succeeded
        assert info.database == athena.database
        assert info.sql == EXAMPLE_SQL
        assert info.state == "SUCCEEDED"

    def test_fail(self, athena):
        query = athena.submit("SELECT x")
        with pytest.raises(AthenaQueryError) as excinfo:
            query.join()
        assert str(excinfo.value).startswith("Athena query failed: SYNTAX_ERROR: ")
        info = query.get_info()
        assert info.finished
        assert not info.succeeded
        assert info.state == "FAILED"
        assert info.database == athena.database
        assert info.sql == "SELECT x"

    def test_kill(self, athena):
        query = athena.submit(EXAMPLE_SQL)
        query.kill()
        with pytest.raises(AthenaQueryError) as excinfo:
            query.join()
        # State change reason is sometimes missing for cancelled queries.
        # Do not assert it here to avoid a flaky test.
        assert str(excinfo.value).startswith("Athena query cancelled")
        info = query.get_info()
        assert info.finished
        assert not info.succeeded
        assert info.state == "CANCELLED"
        assert info.database == athena.database
        assert info.sql == EXAMPLE_SQL

    def test_various_types(self, athena):
        sql = """\
            SELECT
                'anonymous',
                null unknown_null,
                CAST('' as VARCHAR) varchar_empty,
                CAST('b' as VARCHAR) varchar_value,
                CAST(NULL as VARCHAR) varchar_null,
                CAST('a' as CHAR) char_value,
                CAST(NULL as CHAR) char_null,
                true boolean_true,
                false boolean_false,
                cast(null AS BOOLEAN) boolean_null,
                CAST(1 as TINYINT) tinyint_value,
                CAST(2 as SMALLINT) smallint_value,
                CAST(3 as INTEGER) integer_value,
                CAST(4 as BIGINT) bigint_value,
                CAST(null as INTEGER) integer_null,
                CAST(0.1 as REAL) real_value,
                CAST(0.2 as DOUBLE) double_value,
                CAST(null as DOUBLE) double_null,
                nan() double_nan,
                infinity() double_plus_infinity,
                -infinity() double_minus_infinity,
                DECIMAL '0.1' decimal_value,
                CAST(NULL as DECIMAL) decimal_null,
                DATE '2001-08-22' date_value,
                CAST(NULL as DATE) date_null,
                TIMESTAMP '2001-08-22 03:04:05.321' timestamp_value,
                CAST(NULL as TIMESTAMP) timestamp_null,
                from_base64('AP8A') binary_value,
                CAST(NULL as VARBINARY) binary_null,
                ARRAY['item1', 'item2'] array_value,
                CAST(NULL AS ARRAY(VARCHAR)) array_null,
                MAP(ARRAY['k'], ARRAY['v']) map_value,
                CAST(NULL AS MAP(VARCHAR, VARCHAR)) map_null,
                CAST(ARRAY[1, NULL, 456] AS JSON) json_value,
                CAST(NULL AS JSON) json_null
        """
        results = athena.execute(textwrap.dedent(sql))
        assert list(results) == [
            {
                "_col0": "anonymous",
                "unknown_null": None,
                "varchar_empty": "",
                "varchar_value": "b",
                "varchar_null": None,
                "char_value": "a",
                "char_null": None,
                "boolean_true": True,
                "boolean_false": False,
                "boolean_null": None,
                "tinyint_value": 1,
                "smallint_value": 2,
                "integer_value": 3,
                "bigint_value": 4,
                "integer_null": None,
                "real_value": 0.1,
                "double_value": 0.2,
                "double_null": None,
                "double_nan": pytest.approx(float("nan"), nan_ok=True),
                "double_plus_infinity": float("inf"),
                "double_minus_infinity": float("-inf"),
                "decimal_value": Decimal("0.1"),
                "decimal_null": None,
                "timestamp_value": dt.datetime(2001, 8, 22, 3, 4, 5, 321000),
                "timestamp_null": None,
                "binary_value": b"\x00\xff\x00",
                "binary_null": None,
                "date_value": dt.date(2001, 8, 22),
                "date_null": None,
                "array_value": ["item1", "item2"],
                "array_null": None,
                "map_value": {"k": "v"},
                "map_null": None,
                "json_value": [1, None, 456],
                "json_null": None,
            }
        ]

    def test_empty_results(self, athena):
        sql = "SELECT * FROM (VALUES (1, 'a')) AS t (id, name) WHERE id < 0"
        results = athena.execute(sql)
        assert list(results) == []

    def test_long_results(self, athena):
        r1, r2 = range(20), range(100)
        sql = f"""\
            SELECT v1, v2, NULL v3 FROM
                (VALUES {', '.join(map(str, r1))}) AS t1 (v1),
                (VALUES {', '.join(map(str, r2))}) AS t2 (v2)
            ORDER BY
                v1, v2
        """
        results = athena.execute(textwrap.dedent(sql))
        assert list(results) == [
            {"v1": v1, "v2": v2, "v3": None} for v1, v2 in itertools.product(r1, r2)
        ]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime as dt

import pytest

from pallas.exceptions import AthenaQueryError
from pallas.info import QueryInfo, format_price, format_size, format_time


def test_format_price():
    assert format_price(0.00011) == "0.01¢"
    assert format_price(0.0011) == "0.11¢"
    assert format_price(0.011) == "1.10¢"
    assert format_price(0.11) == "11.00¢"
    assert format_price(1.1) == "$1.10"


def test_format_size():
    assert format_size(1.1) == "1B"
    assert format_size(1_100) == "1.10kB"
    assert format_size(1_100_000) == "1.10MB"
    assert format_size(1_100_000_000) == "1.10GB"
    assert format_size(1_100_000_000_000) == "1.10TB"
    assert format_size(1_100_000_000_000_000) == "1100.00TB"


def test_format_time():
    assert format_time(dt.timedelta(seconds=13, milliseconds=570)) == "13.6s"
    assert format_time(dt.timedelta(minutes=13, seconds=57)) == "13min 57s"


EXAMPLE_INFO = {
    "QueryExecutionId": "e83608...1e",
    "Query": "SELECT ...",
    "StatementType": "DML",
    "ResultConfiguration": {
        "OutputLocation": "s3://example-output-location/e83608...1e.csv"
    },
    "QueryExecutionContext": {"Database": "example-database"},
    "Status": {
        "State": "SUCCEEDED",
        "SubmissionDateTime": ...,
        "CompletionDateTime": ...,
    },
    "Statistics": {
        "EngineExecutionTimeInMillis": 300,
        "DataScannedInBytes": 10 ** 9,
        "TotalExecutionTimeInMillis": 400,
        "QueryQueueTimeInMillis": 50,
        "ServiceProcessingTimeInMillis": 20,
    },
    "WorkGroup": "primary",
}


def get_info(*, state=None, state_reason=None):
    data = copy.deepcopy(EXAMPLE_INFO)
    if state is not None:
        data["Status"]["State"] = state
    if state_reason is not None:
        data["Status"]["StateChangeReason"] = state_reason
    return QueryInfo(data)


class TestQueryInfo:
    def test_repr(self):
        info = get_info()
        assert repr(info) == (
            "<QueryInfo: 'SUCCEEDED, scanned 1.00GB in 0.4s, approx. price 0.50¢'>"
        )

    def test_str(self):
        info = get_info()
        assert str(info) == "SUCCEEDED, scanned 1.00GB in 0.4s, approx. price 0.50¢"

    def test_properties(self):
        info = get_info()
        assert info.execution_id == "e83608...1e"
        assert info.sql == "SELECT ..."
        assert info.database == "example-database"
        assert info.output_location == "s3://example-output-location/e83608...1e.csv"
        assert info.scanned_bytes == 1_000_000_000
        assert info.execution_time == dt.timedelta(seconds=0.4)
        assert info.approx_price == 0.005

    @pytest.mark.parametrize("state", ["QUEUED", "RUNNING"])
    def test_running(self, state):
        info = get_info(state=state)
        assert not info.finished
        assert not info.succeeded
        assert info.state == state
        assert info.state_reason is None
        info.check()  # Should not raise

    def test_succeeded(self):
        info = get_info(state="SUCCEEDED")
        assert info.finished
        assert info.succeeded
        assert info.state == "SUCCEEDED"
        assert info.state_reason is None
        info.check()  # Should not raise

    @pytest.mark.parametrize("state", ["FAILED", "CANCELLED"])
    def test_failed(self, state):
        info = get_info(state=state, state_reason="Something happened.")
        assert info.finished
        assert not info.succeeded
        assert info.state == state
        assert info.state_reason == "Something happened."
        with pytest.raises(AthenaQueryError):
            info.check()



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pallas import Athena
from pallas.testing import FakeProxy


@pytest.fixture(name="fake")
def fake_fixture():
    return FakeProxy()


@pytest.fixture(name="athena")
def athena_fixture(fake):
    return Athena(fake)


class TestFakeProxy:
    def test_submit(self, athena, fake):
        query = athena.submit("SELECT ...")
        assert query.execution_id == "query-1"
        assert fake.request_log == ["StartQueryExecution"]

    def test_submit_multiple(self, athena, fake):
        query1 = athena.submit("SELECT ...")
        query2 = athena.submit("SELECT ...")
        assert query1.execution_id == "query-1"
        assert query2.execution_id == "query-2"
        assert fake.request_log == ["StartQueryExecution", "StartQueryExecution"]

    def test_get_query(self, athena):
        query1 = athena.submit("SELECT ...")
        query2 = athena.get_query("query-1")
        assert query2.execution_id == query1.execution_id

    def test_query_info(self, athena, fake):
        query = athena.submit("SELECT ...")
        fake.request_log.clear()
        info = query.get_info()
        assert info.execution_id == "query-1"
        assert info.sql == "SELECT ..."
        assert info.database is None
        assert info.finished
        assert info.succeeded
        assert fake.request_log == ["GetQueryExecution"]

    def test_query_info_remembered(self, athena, fake):
        query = athena.submit("SELECT ...")
        fake.request_log.clear()
        query.get_info()
        query.get_info()
        assert fake.request_log == ["GetQueryExecution"]

    def test_query_info_remembered_not_shared(self, athena, fake):
        query = athena.submit("SELECT ...")
        fake.request_log.clear()
        query.get_info()
        athena.get_query(query.execution_id).get_info()
        assert fake.request_log == ["GetQueryExecution", "GetQueryExecution"]

    def test_default_query_results(self, athena):
        query = athena.submit("SELECT ...")
        results = query.get_results()
        assert list(results) == []

    def test_custom_query_results(self, athena, fake):
        fake.column_names = "id", "name"
        fake.data = [("1", "foo"), ("2", "bar")]
        query = athena.submit("SELECT ...")
        results = query.get_results()
        assert list(results) == [
            {"id": "1", "name": "foo"},
            {"id": "2", "name": "bar"},
        ]

    def test_custom_query_results_with_custom_types(self, athena, fake):
        fake.column_names = "id", "name"
        fake.column_types = "integer", "varchar"
        fake.data = [("1", "foo"), ("2", "bar")]
        query = athena.submit("SELECT ...")
        results = query.get_results()
        assert list(results) == [
            {"id": 1, "name": "foo"},
            {"id": 2, "name": "bar"},
        ]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from .assembly import environ_setup, setup
from .client import Athena, Query
from .exceptions import AthenaQueryError, DatabaseNotFoundError, TableNotFoundError
from .info import QueryInfo
from .results import QueryResults

__all__ = [
    "environ_setup",
    "setup",
    "Athena",
    "Query",
    "QueryInfo",
    "QueryResults",
    "AthenaQueryError",
    "TableNotFoundError",
    "DatabaseNotFoundError",
]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Assorted helpers.
"""

from __future__ import annotations

from typing import Iterator, Optional


class Fibonacci:

    max_value: Optional[int]

    def __init__(self, *, max_value: Optional[int] = None) -> None:
        self.max_value = max_value

    def __iter__(self) -> Iterator[int]:
        a = b = 1
        while self.max_value is None or a < self.max_value:
            yield a
            a, b = b, a + b
        while True:
            yield self.max_value


def truncate_str(v: str, max_length: int = 80) -> str:
    """Trim the given text if too long."""
    if len(v) <= max_length:
        return v
    head = max_length * 2 // 3
    tail = max_length - head - 3
    return v[:head] + "..." + v[-tail:]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Exceptions raised when querying Athena.
"""

from __future__ import annotations

import re
from typing import Optional


class AthenaQueryError(Exception):
    """Athena query failed."""

    #: State of the query execution (FAILED or CANCELLED)
    state: str

    #: Reason of the state of the query execution.
    state_reason: Optional[str]

    def __init__(self, state: str, state_reason: Optional[str]):
        self.state = state
        self.state_reason = state_reason

    def __str__(self) -> str:
        """
        Report query state with its reason.
        """
        if self.state_reason is not None:
            return f"Athena query {self.state.lower()}: {self.state_reason}"
        return f"Athena query {self.state.lower()}"


class DatabaseNotFoundError(AthenaQueryError):
    """
    Athena database does not exist.

    Pallas maps string errors returned by Athena to exception classes.
    """


class TableNotFoundError(AthenaQueryError):
    """
    Athena table does not exist.

    Pallas maps string errors returned by Athena to exception classes.
    """


# Athena does not return structured errors.
# We have to map error messages to exception types.
error_map = [
    # SELECT ...
    (re.compile("Schema (.*) does not exist"), DatabaseNotFoundError),
    # SHOW ...
    (re.compile("Database does not exist: (.*)"), DatabaseNotFoundError),
    # SELECT ...
    (re.compile("Table (.*) does not exist"), TableNotFoundError),
    # INSERT INTO ...
    (re.compile("Table (.*) not found in database (.*)"), TableNotFoundError),
    # SHOW CREATE TABLE ..., SHOW PARTITIONS ..., DESCRIBE TABLE ...
    (re.compile("Table not found (.*)"), TableNotFoundError),
    # SHOW CREATE VIEW ..., DROP VIEW ...
    (re.compile("View not found or not a valid presto view: (.*)"), TableNotFoundError),
]


def get_error(state: str, state_reason: Optional[str]) -> AthenaQueryError:
    error = AthenaQueryError
    if state_reason is not None:
        for pattern, patter_error in error_map:
            if pattern.search(state_reason):
                error = patter_error
                break
    return error(state, state_reason)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helpers for testing.
"""

from __future__ import annotations

from typing import Dict, List, Optional, Sequence

from pallas.info import QueryInfo
from pallas.proxies import AthenaProxy
from pallas.results import QueryResults


class FakeProxy(AthenaProxy):
    """
    Fake of proxy to AWS Athena.

    Can replace :class:`.Boto3Proxy` for testing.
    """

    column_names: Optional[Sequence[str]] = None
    column_types: Optional[Sequence[str]] = None
    state = "SUCCEEDED"
    data: Optional[Sequence[Sequence[str]]] = None

    _sql: Dict[str, str]
    _results: Dict[str, QueryResults]
    _request_log: List[str]

    def __init__(self) -> None:
        self._sql = {}
        self._results = {}
        self._request_log = []

    @property
    def request_log(self) -> List[str]:
        return self._request_log

    def start_query_execution(
        self,
        sql: str,
        database: Optional[str] = None,
        workgroup: Optional[str] = None,
        output_location: Optional[str] = None,
    ) -> str:
        execution_id = f"query-{len(self._results) + 1}"
        self._request_log.append("StartQueryExecution")
        results = self._fake_query_results()
        self._sql[execution_id] = sql
        self._results[execution_id] = results
        return execution_id

    def get_query_execution(self, execution_id: str) -> QueryInfo:
        self._request_log.append("GetQueryExecution")
        return self._fake_query_info(execution_id, self._sql[execution_id])

    def get_query_results(self, info: QueryInfo) -> QueryResults:
        self._request_log.append("GetQueryResults")
        return self._results[info.execution_id]

    def stop_query_execution(self, execution_id: str) -> None:
        self._request_log.append("StopQueryExecution")
        self.state = "CANCELLED"

    def _fake_query_info(self, execution_id: str, sql: str) -> QueryInfo:
        data = {
            "QueryExecutionId": execution_id,
            "Query": sql,
            "ResultConfiguration": {"OutputLocation": ...},
            "QueryExecutionContext": {},
            "Status": {
                "State": self.state,
                "SubmissionDateTime": ...,
                "CompletionDateTime": ...,
            },
            "Statistics": {
                "EngineExecutionTimeInMillis": 0,
                "DataScannedInBytes": 0,
                "TotalExecutionTimeInMillis": 0,
                "QueryQueueTimeInMillis": 0,
                "ServiceProcessingTimeInMillis": 0,
            },
            "WorkGroup": "primary",
        }
        return QueryInfo(data)

    def _fake_query_results(self) -> QueryResults:
        column_names = self.column_names
        column_types = self.column_types
        data = self.data
        if column_names is None:
            column_names = []
        if column_types is None:
            column_types = ["unknown" for _ in column_names]
        if data is None:
            data = []
        return QueryResults(column_names, column_types, data)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages


with open("README.rst", "r") as fp:
    long_description = fp.read()


setup(
    name="pallas",
    version="0.8",
    author="Miloslav Pojman",
    author_email="mpojman@akamai.com",
    description="AWS Athena client",
    keywords=["Amazon", "Athena", "AWS", "database", "db", "Jupyter", "Presto", "SQL"],
    license="Apache License 2.0",
    url="https://github.com/akamai/pallas",
    project_urls={
        "Documentation": "https://pallas.readthedocs.io/",
        "Source Code": "https://github.com/akamai/pallas",
    },
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.7",
    install_requires=["boto3"],
    extras_require={
        "dev": ["flake8", "pytest", "mypy"],
        "pandas": ["pandas>=1.0.0"],
    },
    classifiers=[
        "Framework :: IPython",
        "Framework :: Jupyter",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: SQL",
        "Topic :: Database",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    zip_safe=False,
    include_package_data=True,
)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SQL helpers.

Implements quoting compatible with Athena.
"""

from __future__ import annotations

import base64
import datetime as dt
import math
import numbers
import re
import textwrap
from decimal import Decimal
from typing import List, Mapping, Tuple, Union

SQL_SCALAR = Union[None, str, float, numbers.Real, Decimal, bytes, dt.date]

# Parameters supported for SQL interpolation.
#
# Intentionally use Tuple and List instead of Sequence because
# we do not want to accept strings.
PARAMETERS = Union[
    None, Tuple[SQL_SCALAR, ...], List[SQL_SCALAR], Mapping[str, SQL_SCALAR]
]


def _quote_str(value: str) -> str:
    escaped = str(value).replace("'", "''")
    return f"'{escaped}'"


def _quote_integral(value: Union[int, numbers.Integral]) -> str:
    if not isinstance(value, int):
        value = int(value)
    return str(value)


def _quote_real(value: Union[float, numbers.Real]) -> str:
    if not isinstance(value, float):
        value = float(value)
    if math.isnan(value):
        return "nan()"
    if math.isinf(value):
        return "infinity()" if value > 0 else "-infinity()"
    return str(value)


def _quote_decimal(value: Decimal) -> str:
    return f"DECIMAL {_quote_str(str(value))}"


def _quote_bytes(value: bytes) -> str:
    encoded = base64.b64encode(value).decode("ascii")
    return f"from_base64({_quote_str(encoded)})"


def _quote_datetime(value: dt.datetime) -> str:
    encoded = value.isoformat(sep=" ", timespec="milliseconds")
    return f"TIMESTAMP {_quote_str(encoded)}"


def _quote_date(value: dt.date) -> str:
    encoded = value.isoformat()
    return f"DATE {_quote_str(encoded)}"


def quote(value: SQL_SCALAR) -> str:
    """
    Quote a scalar value for an SQL expression.

    Parametrized queries should be preferred to explicit quoting.

    Following Python types can be quoted to an SQL expressions:

    - :data:`None` – SQL ``NULL``
    - :class:`str`
    - :class:`int`, including subclasses of numbers.Integral
    - :class:`float`, including subclasses or numbers.Real
    - :class:`Decimal` – SQL ``DECIMAL``
    - :class:`datetime.date` – SQL ``DATE``
    - :class:`datetime.datetime` – SQL ``TIMESTAMP``
    - :class:`bytes` – SQL ``VARBINARY``

    :param value: Python value
    :return: an SQL expression
    """
    if value is None:
        return "NULL"
    if isinstance(value, str):
        return _quote_str(value)
    if isinstance(value, int) or isinstance(value, numbers.Integral):
        return _quote_integral(value)
    if isinstance(value, float) or isinstance(value, numbers.Real):
        return _quote_real(value)
    if isinstance(value, Decimal):
        return _quote_decimal(value)
    if isinstance(value, dt.datetime):
        return _quote_datetime(value)
    if isinstance(value, dt.date):
        return _quote_date(value)
    if isinstance(value, bytes):
        return _quote_bytes(value)
    raise TypeError(f"Cannot quote {type(value)}.")


def substitute_parameters(operation: str, parameters: PARAMETERS = None) -> str:
    """
    Substitute parameters in SQL query.
    """
    if parameters is None:
        # When no parameters are given, no substitution happens,
        # so no special quoting is necessary.
        # This is consistent with psycopg2 or MySQLdb behavior.
        return operation
    elif isinstance(parameters, Mapping):
        return operation % {name: quote(param) for name, param in parameters.items()}
    elif isinstance(parameters, (list, tuple)):
        return operation % tuple(quote(param) for param in parameters)
    raise TypeError("SQL parameters must be a sequence or a mapping.")


_comment_1 = r"--[^\n]*\n"
_comment_2 = r"/\*([^*]|\*(?!/))*\*/"

SELECT_RE = re.compile(
    rf"(\s+|{_comment_1}|{_comment_2})*(SELECT|WITH)\b", re.IGNORECASE
)


def is_select(sql: str) -> bool:
    """
    Return whether an SQL statement is SELECT.

    Only SELECT statements are considered cacheable.
    """
    return SELECT_RE.match(sql) is not None


def normalize_sql(sql: str) -> str:
    """
    Normalizes an SQL query.

    Query normalization can improve caching.

    Following normalization operations are done:
    - Common indentation is removed.
    - Heading and trailing new lines are removed.
    - Trailing whitespace is removed from end of lines.
    - Line endings are normalized to LF
    """
    lines = sql.splitlines()
    joined = "\n".join(line.rstrip() for line in lines)
    return textwrap.dedent(joined).strip()



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io

import pytest

from pallas.csv import read_csv, write_csv


def read_csv_str(s):
    stream = io.StringIO(s)
    return list(read_csv(stream))


def write_csv_str(data):
    stream = io.StringIO()
    write_csv(data, stream)
    return stream.getvalue()


class TestReadCSV:
    def test_empty(self):
        assert read_csv_str("") == []

    @pytest.mark.parametrize(
        "raw,value",
        [
            ("", None),
            ('""', ""),
            ('"1"', "1"),
            ('"abc"', "abc"),
            ('""""', '"'),
            ('"foo=""bar"""', 'foo="bar"'),
        ],
    )
    def test_value(self, raw, value):
        assert read_csv_str(f"{raw}\n") == [(value,)]

    @pytest.mark.parametrize("raw", ["1", '"', '"x', '""x'])
    def test_invalid_value(self, raw):
        with pytest.raises(ValueError):
            read_csv_str(f"{raw}\n")

    def test_multiple_columns(self):
        assert read_csv_str('"foo","bar"\n') == [("foo", "bar")]

    def test_multiple_columns_with_none(self):
        assert read_csv_str(",\n") == [(None, None)]

    def test_multiple_rows(self):
        assert read_csv_str('"foo"\n"bar"\n') == [("foo",), ("bar",)]

    def test_multiple_rows_with_none(self):
        assert read_csv_str("\n\n") == [(None,), (None,)]

    def test_long(self):
        value = 10_000 * "abc"
        assert read_csv_str(f'"{value}"\n') == [(value,)]

    def test_long_none_value(self):
        value = 10_000 * "abc"
        assert read_csv_str(f'"{value}",\n') == [(value, None)]

    def test_missing_trailing_newline(self):
        with pytest.raises(ValueError):
            read_csv_str('"foo","bar"')

    def test_missing_trailing_newline_after_none(self):
        with pytest.raises(ValueError):
            read_csv_str(",")


class TestWriteCSV:
    def test_empty(self):
        assert write_csv_str([]) == ""

    @pytest.mark.parametrize(
        "raw,value",
        [
            ("", None),
            ('""', ""),
            ('"1"', "1"),
            ('"abc"', "abc"),
            ('""""', '"'),
            ('"foo=""bar"""', 'foo="bar"'),
        ],
    )
    def test_value(self, raw, value):
        assert write_csv_str([(value,)]) == f"{raw}\n"

    def test_multiple_columns(self):
        assert write_csv_str([("foo", "bar")]) == '"foo","bar"\n'

    def test_multiple_columns_with_none(self):
        assert write_csv_str([(None, None)]) == ",\n"

    def test_multiple_rows(self):
        assert write_csv_str([("foo",), ("bar",)]) == '"foo"\n"bar"\n'

    def test_multiple_rows_with_none(self):
        assert write_csv_str([(None,), (None,)]) == "\n\n"

    def test_long(self):
        value = 10_000 * "abc"
        assert write_csv_str([(value,)]) == f'"{value}"\n'

    def test_long_none_value(self):
        value = 10_000 * "abc"
        assert write_csv_str([(value, None)]) == f'"{value}",\n'



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
from typing import Any, Dict, Optional, TextIO, Tuple
from urllib.parse import urlsplit

import boto3
import botocore

from pallas.storage.base import NotFoundError, Storage, UnsupportedURIError


def s3_parse_uri(uri: str) -> Tuple[str, str]:
    """
    Get bucket name and path from an S3 uri.
    """
    scheme, netloc, path, query, fragment = urlsplit(uri, scheme="s3")
    if scheme != "s3":
        raise UnsupportedURIError("Not a s3 scheme.")
    if query or fragment:
        raise UnsupportedURIError("The scheme does not support query or fragment.")
    if not netloc:
        raise UnsupportedURIError("Bucket is empty.")
    return netloc, path.lstrip("/")


def s3_wrap_body(body: Any) -> TextIO:
    """
    Wrap body returned by S3 GetObject to a valid TextIO
    """
    # Use _raw_stream (urllib3.response.HTTPResponse) because
    # boto3 wrapper does not implement full Binary I/O interface.
    raw = body._raw_stream
    # TextIOWrapper does not work with auto_close
    # https://urllib3.readthedocs.io/en/latest/user-guide.html#using-io-wrappers-with-response-content
    raw.auto_close = False
    return io.TextIOWrapper(raw, encoding="utf-8", newline="")


class S3Storage(Storage):
    """
    Storage implementation storing data in AWS S3.
    """

    _client: Any  # boto3 S3 client

    def __init__(
        self, bucket: str, prefix: str = "", *, client: Optional[Any] = None
    ) -> None:
        if client is None:
            client = boto3.client("s3")
        self._client = client
        self._bucket = bucket
        self._prefix = prefix

    @classmethod
    def from_uri(cls, uri: str) -> S3Storage:
        bucket, prefix = s3_parse_uri(uri)
        if prefix and not prefix.endswith("/"):
            prefix += "/"
        return cls(bucket, prefix)

    @property
    def uri(self) -> str:
        return f"s3://{self._bucket}/{self._prefix}"

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def prefix(self) -> str:
        return self._prefix

    def get(self, key: str) -> str:
        with self.reader(key) as stream:
            return stream.read()

    def set(self, key: str, value: str) -> None:
        params = self._get_params(key)
        params["Body"] = value.encode("utf-8")
        self._client.put_object(**params)

    def has(self, key: str) -> bool:
        params = self._get_params(key)
        try:
            self._client.head_object(**params)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            raise
        return True

    def reader(self, key: str) -> TextIO:
        params = self._get_params(key)
        try:
            response = self._client.get_object(**params)
        except self._client.exceptions.NoSuchKey:
            raise NotFoundError(key) from None
        return s3_wrap_body(response["Body"])

    def _get_params(self, key: str) -> Dict[str, Any]:
        return dict(Bucket=self.bucket, Key=self.prefix + key)
